(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Provers`ComplexityGBBivariate`"]

Needs[ "Theorema`Provers`"]
Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]

Begin["`Private`"]

(* ::Section:: *)
(* Auxiliary Functions *)

(* ::Subsection:: *)
(* Dialogs *)

instantiationDialog[ v_, const_List, {g_, kb_}, k_, title_] :=
    Module[ {expr, 
    		fixBut = Map[ 
    			PasteButton[
    				theoremaDisplay[ RNG$[ #]], 
    				With[ {fbox = ToBoxes[ First[ #], TheoremaForm], fc = First[ #]}, DisplayForm[ InterpretationBox[ fbox, fc]]]]&, const],
    		buttonRow},
        expr[_] = Null;
        buttonRow = {CancelButton[ translate[ "instantiate later"], DialogReturn[ $Failed]],
        			DefaultButton[ translate[ "OK"], DialogReturn[ Array[ expr, Length[ v]]]]};
        tmaDialogInput[
        	Notebook[
	        	Join[
	        		{Cell[ translate[ title], "Subsection"],
	        		Cell[ translate[ "currentGoal"], "Subsubsection"],
	        		Theorema`Provers`Private`goalCell[ g],
	        		Cell[ translate[ "currentKnowledge"], "Subsubsection"]},
	        		Map[ formulaHighlightCells[ #, FML$[ k, _, __]]&, kb],
	        		{Cell[ translate[ "instVar"], "Subsubsection"]},
	        		MapIndexed[ Cell[ BoxData[ RowBox[ {ToBoxes[ #1, TheoremaForm], ":=", 
	        			ToBoxes[ InputField[ Dynamic[ expr[#2[[1]]]], Boxes, FieldSize -> Medium]]}]], "Text"]&, v], 
	        		{Cell[ translate[ "availConst"], "Subsubsection"],
	        		Cell[ BoxData[ RowBox[ Map[ ToBoxes, fixBut]]], "Text"],
	        		Cell[ BoxData[ RowBox[ Map[ ToBoxes, buttonRow]]], "Text"]}
        		]
        	],
        	"Dialog"
        ]
    ]
instantiationDialog[ args___] := unexpected[ instantiationDialog, {args}]

instantiateForallKBInteractive[ const_List, rng_RNG$, {g_FML$, kb_List}, k_] :=
	Module[{vars = rngVariables[ rng], inst, nn},
		inst = instantiationDialog[ vars, const, {g, kb}, k, "instantiateForallHeader"];
		NotebookClose[ $TMAproofNotebook];
		If[ inst === $Failed,
			Return[ $Failed],
			(* else *)
			nn = Position[ inst, Except[ Null], {1}, Heads -> False];
			(* Ignore Null in inst, these come from empty input fields -> these vars should not be instantiated *)
			Block[ {$parseTheoremaExpressions = True, $ContextPath = Join[ {"Theorema`Language`"}, $TheoremaArchives, $ContextPath]},
				(* We need "Theorema`Language`" in the context path already at this point, because otherwise symbols
					like "QU$", "VAR$", etc. would go to "Global`" context. *)
				inst = Map[ ToExpression[ #, StandardForm, Hold]&, Extract[ inst, nn]];
				Begin["Theorema`Knowledge`"];
				inst = Map[ makeTmaExpression, inst];
				End[]
			];
			Thread[ Extract[ vars, nn] -> inst]
		]
	]
instantiateForallKBInteractive[ args___] := unexpected[ instantiateForallKBInteractive, {args}]

instantiateExistsGoalInteractive[ g:FML$[ _, Theorema`Language`Exists$TM[ rng_, __], ___], const_List, kb_List] :=
	Module[{vars = rngVariables[ rng], inst, nn},
		inst = instantiationDialog[ vars, const, {g, kb}, $Failed, "instantiateExistsHeader"];
		If[ inst === $Failed,
			Return[ $Failed],
			(* else *)
			nn = Position[ inst, Except[ Null], {1}, Heads -> False];
			(* Ignore Null in inst, these come from empty input fields -> these vars should not be instantiated *)
			Block[ {$parseTheoremaExpressions = True, $ContextPath = Join[ {"Theorema`Language`"}, $TheoremaArchives, $ContextPath]},
				inst = Map[ ToExpression[ #, StandardForm, Hold]&, Extract[ inst, nn]];
				Begin["Theorema`Knowledge`"];
				inst = Map[ makeTmaExpression, inst];
				End[]
			];
			Thread[ Extract[ vars, nn] -> inst]
		]
	]
instantiateExistsGoalInteractive[ args___] := unexpected[ instantiateExistsGoalInteractive, {args}]

selectAssumptionDialog[ goal_FML$, k:{__FML$}, title_String, cancel_String] :=
 Module[ {proofCells, showProof = False, buttonRow},
    	buttonRow = {CancelButton[ translate[ cancel], DialogReturn[ $Failed]], DefaultButton[ translate[ "OK"], DialogReturn[ $selectedFormula]]};
    	$selectedFormula = 0;
        proofCells = pObjCells[];
        $TMAproofNotebook = tmaNotebookPut[ Notebook[ proofCells], "Proof", Visible -> Dynamic[ showProof]];
        tmaDialogInput[ Notebook[
            Join[ 
                {Cell[ BoxData[ ToBoxes[ 
                    Toggler[ Dynamic[ showProof], 
                        {False -> Tooltip[ translate[ "more"], translate[ "showProofProgress"]], 
                         True -> Tooltip[ translate[ "hide proof"], translate[ "hideProofProgress"]]}]]], 
                    "Hint"],
                 Cell[ translate[ title], "Subsection"],
                 Cell[ translate[ "currentGoal"], "Subsubsection"],
                 Theorema`Provers`Private`goalCell[ goal],
                 Cell[ translate[ "currentKnowledge"], "Subsubsection"]
                },
                MapIndexed[ formulaChoiceButtons, k], 
                {Cell[ BoxData[ RowBox[ Map[ ToBoxes, buttonRow]]], "Text"]}]
            ], "Dialog"]
    ]
selectAssumptionDialog[ goal_FML$, k:{__FML$}, pattern_, title_String, cancel_String] :=
 Module[ {proofCells, showProof = False, buttonRow},
    	buttonRow = {CancelButton[ translate[ cancel], DialogReturn[ $Failed]], DefaultButton[ translate[ "OK"], DialogReturn[ $selectedFormula]]};
    	$selectedFormula = 0;
        proofCells = pObjCells[];
        $TMAproofNotebook = tmaNotebookPut[ Notebook[ proofCells], "Proof", Visible -> Dynamic[ showProof]];
        tmaDialogInput[ Notebook[
            Join[ 
                {Cell[ BoxData[ ToBoxes[ 
                    Toggler[ Dynamic[ showProof], 
                        {False -> Tooltip[ translate[ "more"], translate[ "showProofProgress"]], 
                         True -> Tooltip[ translate[ "hide proof"], translate[ "hideProofProgress"]]}]]], 
                    "Hint"],
                 Cell[ translate[ title], "Subsection"],
                 Cell[ translate[ "currentGoal"], "Subsubsection"],
                 Theorema`Provers`Private`goalCell[ goal],
                 Cell[ translate[ "currentKnowledge"], "Subsubsection"]
                },
                MapIndexed[ formulaChoiceButtons[ #1, pattern, #2]&, k], 
                {Cell[ BoxData[ RowBox[ Map[ ToBoxes, buttonRow]]], "Text"]}]
            ], "Dialog"]
    ]
selectAssumptionDialog[ args___] := unexpected[ exchangeGoalKBDialog, {args}]

formulaChoiceButtons[ f_FML$, {num_Integer}] :=
	Cell[ BoxData[ Theorema`Provers`Private`formulaBox[ f]], "Assumption",
		CellFrameLabels -> {{None, Cell[ makeLabel[ label@f], "AssumptionLabel"]}, {None, None}},
		CellDingbat -> Cell[ BoxData[ ToBoxes[ RadioButton[ Dynamic[ $selectedFormula], num]]], "Assumption"]
	]
formulaChoiceButtons[ f_FML$, pattern_, {num_Integer}] /; MatchQ[ f, pattern] := formulaChoiceButtons[ f, {num}]
formulaChoiceButtons[ f_FML$, _, {_Integer}] :=
	Cell[ BoxData[ Theorema`Provers`Private`formulaBox[ f]], "Assumption",
		CellFrameLabels -> {{None, Cell[ makeLabel[ label@f], "AssumptionLabel"]}, {None, None}}
	]
formulaChoiceButtons[ args___] := unexpected[ formulaChoiceButtons, {args}]

formulaHighlightCells[ f_FML$, pattern_] /; MatchQ[ f, pattern] :=
	Cell[ BoxData[ Theorema`Provers`Private`formulaBox[ f]], "Assumption",
		CellFrameLabels -> {{None, Cell[ makeLabel[ label@f], "AssumptionLabel"]}, {None, None}},
		CellDingbat -> "\[FilledRightTriangle]"
	]
formulaHighlightCells[ f_FML$, _] :=
	Cell[ BoxData[ Theorema`Provers`Private`formulaBox[ f]], "Assumption",
		CellFrameLabels -> {{None, Cell[ makeLabel[ label@f], "AssumptionLabel"]}, {None, None}}
	]
formulaHighlightCells[ args___] := unexpected[ formulaHighlightCells, {args}]


selectGoalDialog[ g:{__FML$}, k:{___FML$}, title_String, cancel_String] :=
 Module[ {proofCells, showProof = False, buttonRow},
    	buttonRow = {CancelButton[ translate[ cancel], DialogReturn[ $Failed]], DefaultButton[ translate[ "OK"], DialogReturn[ $selectedFormula]]};
    	$selectedFormula = 0;
        proofCells = pObjCells[];
        $TMAproofNotebook = tmaNotebookPut[ Notebook[ proofCells], "Proof", Visible -> Dynamic[ showProof]];
        tmaDialogInput[ Notebook[
            Join[ 
                {Cell[ BoxData[ ToBoxes[ 
                    Toggler[ Dynamic[ showProof], 
                        {False -> Tooltip[ translate[ "more"], translate[ "showProofProgress"]], 
                         True -> Tooltip[ translate[ "hide proof"], translate[ "hideProofProgress"]]}]]], 
                    "Hint"],
                 Cell[ translate[ title], "Subsection"],
                 Cell[ translate[ "currentGoal"], "Subsubsection"]
                },
                MapIndexed[ goalChoiceButtons, g],
                {Cell[ translate[ "currentKnowledge"], "Subsubsection"]},
                Map[ formulaHighlightCells[#, Except[ _]]&, k], 
                {Cell[ BoxData[ RowBox[ Map[ ToBoxes, buttonRow]]], "Text"]}]
            ], "Dialog"]
    ]
    
goalChoiceButtons[ f_FML$, {num_Integer}] :=
	Cell[ BoxData[ Theorema`Provers`Private`formulaBox[ f]], "Goal",
		CellFrameLabels -> {{None, Cell[ makeLabel[ label@f], "GoalLabel"]}, {None, None}},
		CellDingbat -> Cell[ BoxData[ ToBoxes[ RadioButton[ Dynamic[ $selectedFormula], num]]], "Goal"]
	]


(* ::Subsection:: *)
(* Rewrite Rules *)

(* ::Subsubsection:: *)
(* Generating Rewrite Rules *)

(*
This subsection contains methods for generating rewrite rules. Most of them can also be found,
for instance, in "FormulaManipulation" in  *slightly different*  versions. Differences include
- All formulae (even logical combinations) are rewritten to their conditions
- Equalities/Equivalences are rewritten only from left to right, regardless of logical quantifiers
- No existential quantifiers are introduced (neither in backward rules nor in conditions)
*)

myFormulaToRules[ orig:FML$[ _, form_, __]] :=
	Module[{stripUniv, ruleList},
		stripUniv = stripUniversalQuantifiers[ form];
		ruleList = Transpose[ Map[ makeRules[ #, orig]&, stripUniv]];
		Map[ Apply[ Join, #]&, ruleList]
	]
formulaToRules[ args___] := unexpected[ formulaToRules, {args}]

(*
The difference between 'stripUniversalQuantifiers' in "FormulaManipulation' and here is the following:
Here, conjunctions are immediately split into several formulae, giving rise to several rewrite rules.
*)
stripUniversalQuantifiers[ Forall$TM[ r_, True, f_]] :=
	Module[ {rc, rv, vars, cond, inner, tmp, out = {}},
		rv = rngVariables[ r];
		rc = myRngToCondition[ r];
		tmp = stripUniversalQuantifiers[ f];
		Scan[(
			{inner, cond, vars} = #;
			AppendTo[ out, {inner, Join[ rc, cond], Join[ rv, vars]}]
			)&,
			tmp
		];
		out
	]
stripUniversalQuantifiers[ Forall$TM[ r_, c_, f_]] :=
	Module[ {rc, rv, vars, cond, inner, tmp, out = {}},
		rv = rngVariables[ r];
		rc = myRngToCondition[ r];
		tmp = stripUniversalQuantifiers[ f];
		Scan[(
			{inner, cond, vars} = #;
			AppendTo[ out, {inner, Theorema`Language`FormulaManipulation`Private`joinConditions[ Join[ rc, cond], c], Join[ rv, vars]}]
			)&,
			tmp
		];
		out
	]
stripUniversalQuantifiers[ Implies$TM[ True, r_]] :=
	stripUniversalQuantifiers[ r]
stripUniversalQuantifiers[ Implies$TM[ l_, r_]] :=
	Module[ {vars, cond, inner, tmp, out = {}},
		tmp = stripUniversalQuantifiers[ r];
		Scan[(
			{inner, cond, vars} = #;
			AppendTo[ out, {inner, Theorema`Language`FormulaManipulation`Private`joinConditions[ cond, l], vars}]
			)&,
			tmp
		];
		out
	]
stripUniversalQuantifiers[ And$TM[ args___]] := Apply[ Join, Map[ stripUniversalQuantifiers, {args}]]
stripUniversalQuantifiers[ expr_] := {{expr, {}, {}}}
stripUniversalQuantifiers[ args___] := unexpected[ stripUniversalQuantifiers, {args}]
	
makeRules[ {(IffDef$TM|EqualDef$TM)[ l_, r_], c_List, var_List}, ref_] := 
	{{}, {makeSingleRule[ {l, r, c, var}, ref]}}
(* Special case equality/equivalence: we can rewrite left to right, but we can also rewrite the equality as a whole. *)
makeRules[ {form:((Equal$TM|Iff$TM)[ l_, r_]), c_List, var_List}, ref_] :=
    {{makeSingleRule[ {form, makeConjunction[ c, And$TM], {}, var}, ref, "backward"]}, {makeSingleRule[ {l, r, c, var}, ref]}}
(* makeRules[ {form:And$TM[ f__], c_List, var_List}, ref_] := 
	{Map[ makeSingleRule[ {#, makeConjunction[ c, And$TM], {}, var}, ref, "backward"]&, {f}], {}} *)
makeRules[ {form_, c:{c1_, cr___}, var_List}, ref_] := 
	{{makeSingleRule[ {form, makeConjunction[ c, And$TM], {}, var}, ref, "backward"]}, {}}
(* Rewrite unconditional formulae to True *)
makeRules[ {form_, {}, var_List}, ref_] :=
	{{makeSingleRule[ {form, True, {}, var}, ref, "backward"]}, {}}
makeRules[ args___] := unexpected[ makeRules, {args}]

(* Do not rewrite numbers *)
makeSingleRule[ {l_?NumberQ, r_, c_List, var_List}, ref_] := Sequence[]
makeSingleRule[ {l_, r_, c_List, var_List}, ref_] /; With[ {fr = freeVariables[ r], fl = freeVariables[ l], fc = freeVariables[ c]}, Complement[ fc, fl] =!= {} || Complement[ fr, fl] =!= {} || Complement[ fl, var] =!= {}] := 
	Sequence[]
makeSingleRule[ {l_, r_, c_List, var_List}, ref_] := 
	{key@ref, mmaTransRule[ {l, r, c, var}, ref]}
makeSingleRule[ {l_, r_, {}, var_List}, ref_, "backward"] := 
	makeSingleRule[ {l, r, {}, var}, ref]
makeSingleRule[ {l_, r_, c_List, var_List}, ref_, "backward"] := 
	makeSingleRule[ { l, makeConjunction[ Append[ c, r], And$TM], {}, var}, ref, "backward"]
makeSingleRule[ args___] := unexpected[ makeSingleRule, {args}]

mmaTransRule[ {l_, r_, c_List, var_List}, ref_] :=
    With[ {left = myExecLeft[ Apply[ Hold, {First[boundVarToPattern[ l, var]]}], var], 
               cond = makeConjunction[ c, And$TM],
               right = myExecRight[ Hold[r], var]},
            ToExpression[ 
            	"RuleDelayed[" <> left <> "," <> 
            	"Sow[" <> ToString[ ref, InputForm] <> ",\"ref\"]; Sow[" <> myExecCondition[ Hold[ cond], var] <> ",\"cond\"];" <> right <> "]"
            ]
        ]
mmaTransRule[ args___] := unexpected[ mmaTransRule, {args}]

myExecLeft[ e_Hold, var_List] := 
	Module[ {s},
		s = substituteFree[ e, Map[ Theorema`Language`FormulaManipulation`Private`varToPattern, var]];
		ReleaseHold[ Map[ Function[ expr, Theorema`Language`FormulaManipulation`Private`toInputString[ Hold[ expr], True], {HoldAll}], s]]
	]
myExecCondition[ e_Hold, var_List] := 
	Module[ {s},
		s = substituteFree[ e, Map[ Theorema`Language`FormulaManipulation`Private`stripVar, var]] /. {Theorema`Language`Assign$TM -> Set,
			Theorema`Language`SetDelayed$TM -> SetDelayed, 
			Theorema`Language`CompoundExpression$TM -> CompoundExpression,
			Theorema`Language`List$TM -> List};
		ReleaseHold[ Map[ Function[ expr, Theorema`Language`FormulaManipulation`Private`toInputString[ Hold[ expr], True], {HoldAll}], s]]
	]
myExecRight[ e_Hold, var_List] := 
	Module[ {s},
		s = substituteFree[ e, Map[ Theorema`Language`FormulaManipulation`Private`stripVar, var]] /. {Theorema`Language`Assign$TM -> Set,
			Theorema`Language`SetDelayed$TM -> SetDelayed, 
			Theorema`Language`CompoundExpression$TM -> CompoundExpression,
			Theorema`Language`List$TM -> List};
		ReleaseHold[ Map[ Function[ expr, Theorema`Language`FormulaManipulation`Private`toInputString[ Hold[ expr], True], {HoldAll}], s]]
	]

(* Some quantifiers (e.g. "Sum") can be used together with subscripts. *)
boundVarToPattern[ q_[ r_RNG$, cond_, sub_, form_], vars_List] :=
	Module[ {qvars = rngVariables[ Hold[ r]], subst, newVars = vars},
		subst = Map[ varToNewPattern[ #, newVars]&, qvars];
		{Apply[ q, joinHold[ substituteFree[ Hold[r], subst], joinHold[ substituteFree[ Hold[cond], subst], joinHold[ substituteFree[ Hold[sub], subst], substituteFree[ Hold[form], subst]]]]], newVars}
	]
boundVarToPattern[ q_[ r_RNG$, cond_, form_], vars_List] :=
	Module[ {qvars = rngVariables[ Hold[ r]], subst, newVars = vars},
		subst = Map[ varToNewPattern[ #, newVars]&, qvars];
		{Apply[ q, joinHold[ substituteFree[ Hold[r], subst], joinHold[ substituteFree[ Hold[cond], subst], substituteFree[ Hold[form], subst]]]], newVars}
	]
(* Some quantifiers (e.g. "Let") don't have a condition. *)
boundVarToPattern[ q_[ r_RNG$, form_], vars_List] :=
	Module[ {qvars = rngVariables[ Hold[ r]], subst, newVars = vars},
		subst = Map[ varToNewPattern[ #, newVars]&, qvars];
		{Apply[ q, joinHold[ substituteFree[ Hold[r], subst], substituteFree[ Hold[form], subst]]], newVars}
	]
boundVarToPattern[ expr:(f_[ args___]), vars_List] :=
	Module[ {newVars = vars, tmp, out = f[]},
		Scan[ (
			tmp = boundVarToPattern[ #, newVars];
			newVars = Last[tmp];
			AppendTo[ out, First[ tmp]];
			)&, 
			expr
		];
		{out, newVars}
	]
boundVarToPattern[ expr_, vars_List] := {expr, vars}
boundVarToPattern[ args___] := unexpected[ boundVarToPattern, {args}]

SetAttributes[ varToNewPattern, HoldRest];
varToNewPattern[ v:VAR$[ SEQ0$[ a_]], oldVars_] :=
	Module[ {name = ToString[ a], i},
		If[ MemberQ[ oldVars, v],
			i = Replace[ Max[ Cases[ oldVars, {v, n_} -> n]] + 1, -Infinity -> 0];
			name = name <> ToString[ i];
			AppendTo[ oldVars, {v, i}],
			AppendTo[ oldVars, v]
		];
		With[ {new = ToExpression[ "SEQ0$" <> name]}, v :> Apply[ Pattern, {new, BlankNullSequence[]}]]	
	]
varToNewPattern[ v:VAR$[ SEQ1$[ a_]], oldVars_] :=
	Module[ {name = ToString[ a], i},
		If[ MemberQ[ oldVars, v],
			i = Replace[ Max[ Cases[ oldVars, {v, n_} -> n]] + 1, -Infinity -> 0];
			name = name <> ToString[ i];
			AppendTo[ oldVars, {v, i}],
			AppendTo[ oldVars, v]
		];
		With[ {new = ToExpression[ "SEQ1$" <> name]}, v :> Apply[ Pattern, {new, BlankSequence[]}]]	
	]
varToNewPattern[ v:VAR$[ a_], oldVars_] :=
	Module[ {name = ToString[ a], i},
		If[ MemberQ[ oldVars, v],
			i = Replace[ Max[ Cases[ oldVars, {v, n_} -> n]] + 1, -Infinity -> 0];
			name = name <> ToString[ i];
			AppendTo[ oldVars, {v, i}],
			AppendTo[ oldVars, v]
		];
		With[ {new = ToExpression[ "VAR$" <> name]}, v :> Apply[ Pattern, {new, Blank[]}]]	
	]

trimFormulaForRewriting[ form_FML$] :=
    Module[ {sRules = {}, dRules = {}, goalRules = {}, kbForm = {}},
        Switch[ form,
            (* FML$[ _, (Equal$TM|Iff$TM)[ lhs_, rhs_], __],
            sRules = Apply[ Join, myFormulaToRules[ form]], *)
            (* explicit definitions -> rewrite rules *)
            FML$[ _, _?(!FreeQ[ #, _IffDef$TM|EqualDef$TM[ _, _]]&), __],
            dRules = Last[ myFormulaToRules[ form]],
            _,
            {goalRules, sRules} = myFormulaToRules[ form];
            AppendTo[ kbForm, form]
            ];
        {kbForm, {}, goalRules, sRules, dRules}
    ]

(* ::Subsubsection:: *)
(* Applying Rewrite Rules*)

(* "replaceFirstPossible" behaves very similar to "replaceAllAndTrack", except that it does not generate
	any conditions that have to be proved afterwards, but only applies the first rewrite rule whose conditions
	follow immediately from the KB (It then also returns the formulae which are needed to establish the immediate
	consequence). Function "followsImmediately" is used for this.
	'First rewrite rule' means: Try to rewrite whole expression using the rewrite rules one after the other;
	If this can be done with a rule whose conditions follow immediately, we are done. Otherwise apply
	this function recursively on all subexpressions in breadth-first manner.
	The optional argument 'AcceptTrue' does the following: If True, the result of the function might be an
	expression whose truth value is True; If False, the result will either have truth value different from True,
	or is the original expression. This feature is needed to prevent formulae in the KB from being rewritten into
	True.
	The optional argument 'AcceptExisting' does the following: If True, the result of the function might be an
	expression that is a member of 'k'; If False, the result will either be a new expression not occuring in 'k',
	or the original one.
*)
Options[ replaceFirstPossible] = {AcceptTrue -> True, Except -> {}, Equality -> Weak, ExceptHeads -> {}};
replaceFirstPossible[ expr_, {}, rules_List, ___?OptionQ] := {expr, {}, {}}
replaceFirstPossible[ expr_, repl_List, rules_List, opts___?OptionQ] :=
	Module[ {pos, new, uc, at, eq, ex, eh, sub, condrefs, tmp},
		{at, eq, ex, eh} = {AcceptTrue, Equality, Except, ExceptHeads} /. {opts} /. Options[ replaceFirstPossible];
		pos = Sort[ Position[ expr /. Join[ {(VAR$|SEQ0$|SEQ1$)[_] -> VAR$, FIX$[___] -> FIX$}, Map[ Rule[ #[ ___], #]&, eh]], Except[ VAR$], Heads -> True]];
		(* Built-in "Sort" automatically sorts the positions in breadth-first order. *)
		Catch[
			Scan[
				Function[ p,
					sub = myExtract[ expr, p];
					If[ !MatchQ[ sub, _VAR$|_SEQ0$|_SEQ1$],
						Scan[
							Function[ r,
								{new, uc} = Reap[ Replace[ sub, r], {"ref", "cond"}];
								If[ uc =!= {{}, {}},
									new = myReplacePart[expr, p -> new, Heads -> True];
									If[ (at || !TrueQ[ computeTruthValue[ new]]) && !memberOfKB[ ex, new, Equality -> eq],
										condrefs = {};
										Scan[
											Function[ c,
												tmp = followsImmediately[ c, rules];
												If[ ListQ[ tmp],
													condrefs = Union[ condrefs, tmp],
													condrefs = False;
													Return[];
												]
											],
											uc[[2, 1]]
										];
										(*
										cond = simplifiedAnd[ makeConjunction[ uc[[2,1]], And$TM]];
										cond = If[ TrueQ[ cond], {}, followsImmediately[ cond, rules]];
										*)
										If[ ListQ[ condrefs], Throw[ {new, uc[[1,1]], condrefs}]]
									]
								]
							],
							repl
						]
					]
				],
				pos
			];
			{expr, {}, {}}
		]
	]

(* ::Subsection:: *)
(* General Purpose *)

(* The position specification of the entire expression, as returned by "Position", is {}.
	This, however, cannot be used with functions "Extract" and "ReplacePart", so we define it ourselves. *)
myExtract[ expr_, {}] := expr
myExtract[ expr_, pos_List] := Extract[ expr, pos]
myReplacePart[ _, (Rule|RuleDelayed)[ {}, new_], ___?OptionQ] := new
myReplacePart[ expr_, repl:(_Rule|_RuleDelayed), opts___?OptionQ] := ReplacePart[ expr, repl, opts]

newABF[ name_, formulae_List] :=
	FIX$[ name, (Max[ Cases[ formulae, FIX$[ name, n_] -> n, Infinity]] + 1) /. -Infinity -> 0]

sinkNegation[ Not$TM[ True]] := False
sinkNegation[ Not$TM[ False]] := True
sinkNegation[ Not$TM[ Not$TM[ a_]]] := sinkNegation[ a]
sinkNegation[ Not$TM[ And$TM[ a__]]] := sinkNegation[ Map[ Not$TM, Or$TM[ a]]]
sinkNegation[ Not$TM[ Or$TM[ a__]]] := sinkNegation[ Map[ Not$TM, And$TM[ a]]]
sinkNegation[ Not$TM[ Implies$TM[ a_, b_]]] := And$TM[ a, sinkNegation[ Not$TM[ b]]]
sinkNegation[ Not$TM[ Iff$TM[ a_, b_]]] := Or$TM[ And$TM[ a, sinkNegation[ Not$TM[ b]]], And$TM[ sinkNegation[ Not$TM[ a]], b]]
sinkNegation[ Not$TM[ Forall$TM[ r_RNG$, cond_, f_]]] := Exists$TM[ r, sinkNegation[ cond], sinkNegation[ Not$TM[ f]]]
sinkNegation[ Not$TM[ Exists$TM[ r_RNG$, cond_, f_]]] := Forall$TM[ r, sinkNegation[ cond], sinkNegation[ Not$TM[ f]]]
sinkNegation[ a:(_And$TM|_Or$TM)] := Map[ sinkNegation, a]
sinkNegation[ a:((Implies$TM|Iff$TM|IffDef$TM)[ _, _])] := Map[ sinkNegation, a]
sinkNegation[ (h:Forall$TM|Exists$TM)[ r_RNG$, cond_, f_]] := h[ r, sinkNegation[ cond], sinkNegation[ f]]
sinkNegation[ a_] := a

wEqual[ a_, a_] := True
wEqual[ a_FML$, b_FML$] := wEqual[ formula@a, formula@b]
wEqual[ h_[ r_RNG$, a__], h_[ s_RNG$, b__]] :=
	Module[ {rules, anew},
		If[ Length[ Hold[ a]] === Length[ Hold[ b]] && Length[ r] === Length[ s],
			rules = MapThread[ (#1 -> #2)&, {rngVariables[ r], rngVariables[ s]}];
			anew = Map[ substituteFree[ #, rules]&, {r, a}];
			Apply[ And, MapThread[ wEqual, {anew, {s, b}}]],
			False
		]
	]
wEqual[ g_[ a___], h_[ b___]] := Length[ Hold[ a]] === Length[ Hold[ b]] && wEqual[ g, h] && Apply[ And, MapThread[ wEqual, {{a}, {b}}]]
wEqual[ _, _] := False

Options[ addNewKB] = {Equality -> Weak, Mode -> Prepend};
addNewKB[ kb:{___FML$}, FML$[ _, True|Not$TM[ False], ___], opts___?OptionQ] := kb
addNewKB[ kb:{___FML$}, fml_FML$, opts___?OptionQ] /; $autoGenerateRules :=
	Module[ {member, trimmed, eq, m},
		{eq, m} = {Equality, Mode} /. {opts} /. Options[ addNewKB];
		If[ eq === Strong,
			member = Catch[ Scan[ If[ formula@fml === formula@#, Throw[ True]]&, kb]; False],
			member = Catch[ Scan[ If[ wEqual[ formula@fml, formula@#], Throw[ True]]&, kb]; False]
		];
		If[ member,
			kb,
			(* else *)
			trimmed = trimFormulaForRewriting[ fml];
			AppendTo[ $rewriteRules, Rest[ trimmed]];
			If[ m === Append,
				Join[ kb, First[ trimmed]],
				Join[ First[ trimmed], kb]
			]
		]
	]
addNewKB[ kb:{___FML$}, fml_FML$, opts___?OptionQ] :=
	Module[ {member, eq, m},
		{eq, m} = {Equality, Mode} /. {opts} /. Options[ addNewKB];
		If[ eq === Strong,
			member = Catch[ Scan[ If[ formula@fml === formula@#, Throw[ True]]&, kb]; False],
			member = Catch[ Scan[ If[ wEqual[ formula@fml, formula@#], Throw[ True]]&, kb]; False]
		];
		If[ member,
			kb,
			m[ kb, fml]
		]
	]
addNewKB[ args___] := unexpected[ addNewKB, {args}]

joinNewKB[ kb:{___FML$}, {}, ___?OptionQ] := kb
joinNewKB[ kb:{___FML$}, new:{__FML$}, opts___?OptionQ] :=
	Module[ {m},
		m = Mode /. {opts} /. Options[ addNewKB];
		If[ m === Append,
			Fold[ addNewKB[ #1, #2, opts]&, kb, new],
			Fold[ addNewKB[ #1, #2, opts]&, kb, Reverse[ new]]
		]
	]
joinNewKB[ args___] := unexpected[ joinNewKB, {args}]

memberOfKB[ kb:{___FML$}, fml_FML$, opts___?OptionQ] :=
	Module[ {eq},
		eq = Equality /. {opts} /. Options[ addNewKB];
		If[ eq === Strong,
			Catch[ Scan[ If[ formula@fml === formula@#, Throw[ True]]&, kb]; False],
			Catch[ Scan[ If[ wEqual[ formula@fml, formula@#], Throw[ True]]&, kb]; False]
		]
	]
memberOfKB[ kb:{___FML$}, f_, opts___?OptionQ] :=
	Module[ {eq},
		eq = Equality /. {opts} /. Options[ addNewKB];
		If[ eq === Strong,
			Catch[ Scan[ If[ f === formula@#, Throw[ True]]&, kb]; False],
			Catch[ Scan[ If[ wEqual[ f, formula@#], Throw[ True]]&, kb]; False]
		]
	]

simplifyFormulaByComputation[ FML$[ key_, f_, rest___]] := FML$[ key, computeInProof[ f], rest]
	
(* "extractFormulaFromRule" takes a rule and tries to extract the formula the rule originated from by
	looking at subexpressions of the form "Sow[ fml_, "ref"]". If it succeeds, it returns the formula,
	otherwise it returns $Failed. *)
extractFormulaFromRule[ {_, r:(_Rule|_RuleDelayed), ___}] := extractFormulaFromRule[ r]
extractFormulaFromRule[ (Rule|RuleDelayed)[ _, r_]] :=
	Module[ {list},
		list = Cases[ Hold[ r], HoldPattern[ Sow[ fml:FML$[ _, _, __], "ref"]] :> fml, Infinity];
		If[ Length[ list] === 1,
			First[ list],
			$Failed
		]
	]
	
(* "followsImmediately" is a method for finding out whether a certain formula follows 'immediately' from
	a given knowledge base / backward rules. It is mainly used in inference rules for rewriting when some
	rewrite rules have conditions; If the conditions follow 'immediately', the rewrite rule can be applied,
	otherwise not. *)
followsImmediately[ FML$[ _, f_, __], rules_List] :=
	followsImmediately[ f, rules]
followsImmediately[ And$TM[ args___], rules_List] :=
	Module[ {res = {}, tmp},
		Catch[
			Scan[(
				tmp = followsImmediately[ #, rules];
				If[ tmp === False,
					Throw[ False],
					res = Join[ res, tmp]
				]
				)&,
				{args}
			];
			res
		]
	]
followsImmediately[ f_, rules_List] :=
	Module[ {search},
		search = exhaustiveSearch[ f, True, rules, computeTruthValue, 3];
		If[ MatchQ[ search, {True, _List}],
			Last[ search],
			False
		]
	]
	
(* "computeTruthValue" provides a very elementary mechanism to compute the truth value of a formula. *)
computeTruthValue[ Not$TM[ True]] := False
computeTruthValue[ Not$TM[ False]] := True
computeTruthValue[ And$TM[ True...]] := True
computeTruthValue[ And$TM[ ___, False, ___]] := False
computeTruthValue[ Or$TM[ False...]] := False
computeTruthValue[ Or$TM[ ___, True, ___]] := True
computeTruthValue[ Implies$TM[ a_, a_]] := True
computeTruthValue[ Implies$TM[ False, _]] := True
computeTruthValue[ Implies$TM[ _, True]] := True
computeTruthValue[ Implies$TM[ True, False]] := False
computeTruthValue[ (Iff$TM|IffDef$TM)[ a_, a_]] := True
computeTruthValue[ (Iff$TM|IffDef$TM)[ False, True]] := False
computeTruthValue[ (Iff$TM|IffDef$TM)[ True, False]] := False
computeTruthValue[ (Equal$TM|EqualDef$TM)[ a_, a_]] := True
computeTruthValue[ Forall$TM[ _, a_, a_]] := True
computeTruthValue[ Forall$TM[ _, False, _]] := True
computeTruthValue[ Forall$TM[ _, _, True]] := True
computeTruthValue[ Exists$TM[ _, False, _]] := False
computeTruthValue[ Exists$TM[ _, _, False]] := False
computeTruthValue[ f:((Not$TM|And$TM|Or$TM|Implies$TM|Iff$TM|IffDef$TM|Forall$TM|Exists$TM)[ args___])] :=
	Module[ {res},
		computeTruthValue[ res] /; (res = Map[ computeTruthValue, f]) =!= f
	]
computeTruthValue[ f_] := f
	
exhaustiveSearch[ start_, target_, rules_List, simp_,  maxDepth_Integer] :=
	exhaustiveSearch[ target, rules, simp, {{start, {}}}, {}, maxDepth]
exhaustiveSearch[ target_, _List, _, {___, l:{target_, _List}, ___}, _List, _] := l
exhaustiveSearch[ target_, rules_List, simp_, l:{{first_, refs_List}, rest___}, visited_List, maxDepth_Integer] /; Length[ refs] <= maxDepth :=
	Module[ {new, newExpressions = {}, newNodes = {}, i, expri},
		new = DeleteCases[ Map[ rewriteAllPositiveParts[ first, #, visited]&, rules], {Except[ _List], {}}];
		For[ i = 1, i <= Length[ new], i++,
			expri = Last[ new[[i]]];
			expri = DeleteDuplicates[ Join[ expri, Select[ Map[ simp, expri], !MemberQ[ visited, #]&]]];
			newExpressions = Join[ newExpressions, expri];
			Scan[ (AppendTo[ newNodes, {#, Append[ refs, First[ new[[i]]]]}])&, expri];
		];
		exhaustiveSearch[ target, rules, simp, Join[ {rest}, newNodes], Join[ visited, newExpressions], maxDepth]
	]
exhaustiveSearch[ _, _List, _, _List, _List, _] := $Failed
	
(* "rewriteAllParts" rewrites all parts (except VAR$, SEQ0$, SEQ1$ and inside FIX$) using the given rule.
	It returns a pair consisting of
	- the formula the rule originated from, and
	- a list of all new (rewritten) expressions which are not among 'exceptions'.
*)
rewriteAllParts[ expr_, rule_, exceptions_List] :=
	Module[ {pos, newExpr, ref, cond, out = {}, tmp},
		pos = Position[ expr /. {(VAR$|SEQ0$|SEQ1$)[_] -> VAR$, FIX$[___] -> FIX$}, Except[ VAR$], Heads -> True];
		Scan[(
			tmp = myExtract[ expr, #];
			{newExpr, {ref, cond}} = Reap[ Replace[ tmp, rule], {"ref", "cond"}];
			If[ Hold[ newExpr] =!= Hold[ tmp] && cond === {{True}},
				tmp = myReplacePart[ expr, # -> newExpr];
				If[ !MemberQ[ exceptions, tmp], AppendTo[ out, myReplacePart[ expr, # -> newExpr]]]
			]
			)&,
			pos
		];
		(* if ref =!= {}, then extractFormulaFromRule[ rule] === ref[[1, 1]] *)
		{If[ ref === {}, extractFormulaFromRule[ rule], ref[[1, 1]]], out}
	]
	
(* "rewriteAllPositiveParts" is very similar to "rewriteAllParts", only that it rewrites only the positive
	parts of the expressions, i.e. those occuring inside an even number of Not$TM (also hidden ones, e.g.
	left-hand-side of Implies$TM). The whole expression is assumed to be positive, i.e. without an "invisible"
	Not$TM wrapped around.
*)
rewriteAllPositiveParts[ expr_, rule_, exceptions_List] :=
	Module[ {pos, newExpr, ref, cond, out = {}, tmp},
		pos = positivePositions[ expr /. {(VAR$|SEQ0$|SEQ1$)[_] -> VAR$, FIX$[___] -> FIX$}, Except[ VAR$], {}, True];
		Scan[(
			tmp = myExtract[ expr, #];
			{newExpr, {ref, cond}} = Reap[ Replace[ tmp, rule], {"ref", "cond"}];
			If[ Hold[ newExpr] =!= Hold[ tmp] && cond === {{True}},
				tmp = myReplacePart[ expr, # -> newExpr];
				If[ !MemberQ[ exceptions, tmp], AppendTo[ out, myReplacePart[ expr, # -> newExpr]]]
			]
			)&,
			pos
		];
		(* if ref =!= {}, then extractFormulaFromRule[ rule] === ref[[1, 1]] *)
		{If[ ref === {}, extractFormulaFromRule[ rule], ref[[1, 1]]], out}
	]

positivePositions[ expr:(Not$TM[ A_]), pattern_, {path___}, True] :=
	Module[ {out = positivePositions[ A, pattern, {path, 1}, False]},
		If[ MatchQ[ expr, pattern], AppendTo[ out, {path}]];
		out
	]
positivePositions[ expr:(Not$TM[ A_]), pattern_, {path___}, False] :=
	positivePositions[ A, pattern, {path, 1}, True]
positivePositions[ expr:((And$TM|Or$TM)[ args__]), pattern_, path_List, True] :=
	Module[ {out = Apply[ Join, MapIndexed[ positivePositions[ #1, pattern, Append[ path, First[ #2]], True]&, {args}]]},
		If[ MatchQ[ expr, pattern], AppendTo[ out, path]];
		out
	]
positivePositions[ expr:((And$TM|Or$TM)[ args__]), pattern_, {path___}, False] :=
	Apply[ Join, MapIndexed[ positivePositions[ #1, pattern, {path, First[ #2]}, False]&, {args}]]
positivePositions[ expr:(Implies$TM[ l_, r_]), pattern_, path_List, True] :=
	Module[ {out = Join[ positivePositions[ l, pattern, Append[ path, 1], False], positivePositions[ r, pattern, Append[ path, 2], True]]},
		If[ MatchQ[ expr, pattern], AppendTo[ out, path]];
		out
	]
positivePositions[ expr:(Implies$TM[ l_, r_]), pattern_, {path___}, False] :=
	Join[ positivePositions[ l, pattern, {path, 1}, True], positivePositions[ r, pattern, {path, 2}, False]]
positivePositions[ expr:(Forall$TM[ _, cond_, A_]), pattern_, path_List, True] :=
	Module[ {out = Join[ positivePositions[ cond, pattern, Append[ path, 2], False], positivePositions[ A, pattern, Append[ path, 3], True]]},
		If[ MatchQ[ expr, pattern], AppendTo[ out, path]];
		out
	]
positivePositions[ expr:(Forall$TM[ _, cond_, A_]), pattern_, {path___}, False] :=
	Join[ positivePositions[ cond, pattern, {path, 2}, True], positivePositions[ A, pattern, {path, 3}, False]]
positivePositions[ expr:(Exists$TM[ _, cond_, A_]), pattern_, path_List, True] :=
	Module[ {out = Join[ positivePositions[ cond, pattern, Append[ path, 2], True], positivePositions[ A, pattern, Append[ path, 3], True]]},
		If[ MatchQ[ expr, pattern], AppendTo[ out, path]];
		out
	]
positivePositions[ expr:(Exists$TM[ _, cond_, A_]), pattern_, {path___}, False] :=
	Join[ positivePositions[ cond, pattern, {path, 2}, False], positivePositions[ A, pattern, {path, 3}, False]]
positivePositions[ expr_, pattern_, path_List, pos_] :=
	If[ pos && MatchQ[ expr, pattern], {path}, {}]
	

(* The following functions are used in goal rewriting to obtain only the "simplest" formulae *)
subFormula[ a_, a_] := True
subFormula[ a_And$TM, b_And$TM] := Complement[ a, b] === And$TM[]
subFormula[ a_And$TM, b_] := subFormula[ a, And$TM[ b]]
subFormula[ a_, b_And$TM] := subFormula[ And$TM[ a], b]
subFormula[ a_, b_] := subFormula[ And$TM[ a], And$TM[ b]]

simplestFormulae[ {}] := {}
simplestFormulae[ {a:{_FML$, _}, b:({_FML$, _}...)}] := simplestFormulae[ {a}, {b}]
simplestFormulae[ l_List, {a_, b___}] :=
	simplestFormulae[ updateSimplest[ l, a, {}, True], {b}]
simplestFormulae[ l_List, {}] := l

updateSimplest[ {s:{FML$[ _, f_, __], _}, r___}, t:{FML$[ _, a_, __], _}, new:{n___}, add_] :=
	If[ TrueQ[ subFormula[ a, f]],
		updateSimplest[ {r}, t, new, add],	(* s must be removed *)
		If[ TrueQ[ subFormula[ f, a]],
			updateSimplest[ {r}, t, {n, s}, False],	(* s must be kept, t must not be added *)
			updateSimplest[ {r}, t, {n, s}, add]	(* s must be kept *)
		]
	]
updateSimplest[ {}, _, l_List, False] := l
updateSimplest[ {}, a_, {n___}, True] := {n, a}


existentialPattern[ form_, vars_List] :=
	Module[ {fv = Intersection[ freeVariables[ form], vars]},
		With[ {left = ToExpression[ myExecLeft[ Apply[ Hold, {First[boundVarToPattern[ form, fv]]}], fv]], 
               right = Map[ Theorema`Language`FormulaManipulation`Private`stripVar, fv]},
            RuleDelayed[ left, right]
        ]
	]
	
simpleUnify[ l1_List, l2_List] :=
	DeleteDuplicates[
		DeleteCases[
			Map[ Sort[ DeleteDuplicates[ Apply[ Join, #]]]&, Tuples[ {l1, l2}]],
			{___, x_ -> _, ___, x_ -> _, ___}
		]
	]
	
(* "myRngToCondition" treats step ranges differently: It always assumes that both endpoints, as well as the
	step size, are integers, and gives as corresponding range condition a domain decision predicate. *)
myRngToCondition[ RNG$[ r__]] := Apply[ Join, Map[ myRngToCondition, {r}]]
myRngToCondition[ SIMPRNG$[ v_]] := {}
myRngToCondition[ SETRNG$[ v_, S_]] := {Element$TM[ v, S]}
myRngToCondition[ STEPRNG$[ v_, l_, h_, 1]] := 
	{DomainOperation$TM[ IntegerInterval$TM[ l, h, True, True], Element$TM][ v]}
myRngToCondition[ STEPRNG$[ v_, l_, h_, s_]] := 
	Module[ {new, step},
		step = Times$TM[ new, s];
		{Exists$TM[ RNG$[ SETRNG$[ new, IntegerInterval$TM[ 0, Infinity, True, False]]], True, 
			And$TM[ Equal$TM[ v, Plus$TM[ l, step]],
				If[ TrueQ[ Negative[ s]], GreaterEqual$TM, LessEqual$TM][ v, h]]]}
	]
myRngToCondition[ PREDRNG$[ v_, P_]] := {P[ v]}
myRngToCondition[ u_] := {$Failed}
myRngToCondition[ args___] := unexpected[ myRngToCondition, {args}]

(* "SublistQ" takes two lists 'a' and 'b' and checks whether 'a' is a sublist of 'b', meaning that
	every item of 'a' occurs at least as often in 'b' as it occurs in 'a', no matter at which positions. *)
SublistQ[ a_List, b_List] :=
	Catch[
		Scan[ If[ Count[ b, #] < Count[ a, #], Throw[ False]]&, DeleteDuplicates[ a]];
		True
	]
	
makeConditionFML[ data___?OptionQ] :=
	Module[ {l, form, newLabel},
		l = label /. {data} /. Options[ makeFML];
		newLabel = If[ StringMatchQ[ l, "C"~~"\[NumberSign]"|"\[SpaceIndicator]"~~__],
			(* If label already has the condition marker, then don't add it once more *)
			l,
			(* else: add condition marker *) 
			With[ {sep = If[ StringMatchQ[ l, NumberString], "\[NumberSign]", "\[SpaceIndicator]"]}, "C" <> sep <> l]
		];
		form = makeFML[ label -> newLabel, data];
		AppendTo[ $generated, form];
		form
	]
makeConditionFML[ args___] := unexpected[ makeConditionFML, {args}]

	
(* ::Subsection:: *)
(* Special Purpose *)

elementsOfDomain[ kb:{___FML$}, dom_, (List|DomainOperation$TM[ dom_, Plus$TM])[elements___]] := Apply[ And, Map[ elementsOfDomain[ kb, dom, #]&, Hold[ elements]]]
elementsOfDomain[ _, dom_, DomainOperation$TM[ dom_, 0]] := True
elementsOfDomain[ kb:{___FML$}, dom_, element_] := MemberQ[ kb, FML$[ _, DomainOperation$TM[ dom, Element$TM][ element], __]]

getAllAtomicArguments[ a:(DomainOperation$TM[ dom_, Plus$TM][ args___])] :=
	Apply[ List, Flatten[ a, Infinity, DomainOperation$TM[ dom, Plus$TM]]]
getAllAtomicArguments[ a_] := {a}
getAllAtomicArguments[ a:(DomainOperation$TM[ dom_, Plus$TM][ args___]), dom_, False|PatternSequence[]] :=
	Apply[ List, Flatten[ a, Infinity, DomainOperation$TM[ dom, Plus$TM]]]
getAllAtomicArguments[ a_, _, False|PatternSequence[]] := {a}
getAllAtomicArguments[ a:(DomainOperation$TM[ dom_, Plus$TM][ args___]), dom_, True] :=
	DeleteCases[ Apply[ List, Flatten[ a, Infinity, DomainOperation$TM[ dom, Plus$TM]]], DomainOperation$TM[ dom, 0]]
getAllAtomicArguments[ a_, dom_, True] := DeleteCases[ {a}, DomainOperation$TM[ dom, 0]]

getFirstUnusedTransitivityFormulae[ KB:{___, k:FML$[ _, DomainOperation$TM[ dom_, r1:(Less$TM|LessEqual$TM)][ x_, y_], __], ___, c:FML$[ _, DomainOperation$TM[ dom_, r2:(Less$TM|LessEqual$TM)][ y_, z_], __], ___}, used_List] :=
	{k, c, dom, x, y, z, r1, r2} /; (
		x =!= y && y =!= z &&
		MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTotallyOrdered$TM|Theorema`Knowledge`isTOM$TM)[ dom], __]] &&
		!MemberQ[ used, {key@k, key@c}|{key@c, key@k}] &&
		If[ x === z,
			!MemberQ[ KB, FML$[ _, (Equal$TM[ x, y])|(Equal$TM[ y, x]), __]],
			!MemberQ[ KB, FML$[ _, DomainOperation$TM[ dom, If[ SameQ[ r1, r2, LessEqual$TM], LessEqual$TM, Less$TM]][ x, z], __]]
		]
	)
getFirstUnusedTransitivityFormulae[ KB:{___, c:FML$[ _, DomainOperation$TM[ dom_, r2:(Less$TM|LessEqual$TM)][ y_, z_], __], ___, k:FML$[ _, DomainOperation$TM[ dom_, r1:(Less$TM|LessEqual$TM)][ x_, y_], __], ___}, used_List] :=
	{k, c, dom, x, y, z, r1, r2} /; (
		x =!= y && y =!= z &&
		MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTotallyOrdered$TM|Theorema`Knowledge`isTOM$TM)[ dom], __]] &&
		!MemberQ[ used, {key@k, key@c}|{key@c, key@k}] &&
		If[ x === z,
			!MemberQ[ KB, FML$[ _, (Equal$TM[ x, y])|(Equal$TM[ y, x]), __]],
			!MemberQ[ KB, FML$[ _, DomainOperation$TM[ dom, If[ SameQ[ r1, r2, LessEqual$TM], LessEqual$TM, Less$TM]][ x, z], __]]
		]
	)
getFirstUnusedTransitivityFormulae[ _List, _List] := {$Failed, $Failed, $Failed, $Failed, $Failed, $Failed, $Failed, $Failed}


acuNormalize[ expr_, kb_List] :=
	Module[ {tmp, pos, dom, term, new, out = expr},
		tmp = expr /. {DomainOperation$TM[ d_, Plus$TM][ ___] :> "$+$" /; MemberQ[ kb, FML$[ _, (Theorema`Knowledge`isAbelianMonoid$TM|Theorema`Knowledge`isTOM$TM)[ d], __]]};
		pos = Position[ tmp, "$+$"];
		Scan[ Function[ p,
				term = myExtract[ expr, p];
				dom = First[ Head[ term]];
				new = term //. {
									DomainOperation$TM[ dom, Plus$TM][ ] :> DomainOperation$TM[ dom, 0],
									DomainOperation$TM[ dom, Plus$TM][ x_] :> x /; elementsOfDomain[ kb, dom, x],
									a:(DomainOperation$TM[ dom, Plus$TM][ pre___, DomainOperation$TM[ dom, 0], post___]) :> DeleteCases[ a, DomainOperation$TM[ dom, 0]] /; elementsOfDomain[ kb, dom, {pre, post}],
									DomainOperation$TM[ dom, Plus$TM][ pre___, DomainOperation$TM[ dom, Plus$TM][ args___], post___] :> DomainOperation$TM[ dom, Plus$TM][ pre, args, post] /; elementsOfDomain[ kb, dom, {pre, args, post}],
									a:(DomainOperation$TM[ dom, Plus$TM][ args___]) :> Sort[ a] /; (!OrderedQ[ a] && elementsOfDomain[ kb, dom, {args}])
								};
				If[ new =!= term,
					out = myReplacePart[ out, p -> new]
				]
			],
			pos
		];
		out
	]

(* ::Section:: *)
(* Propositional Logic *)

(* The following rules are almost exactly the same as the ones in "BasicTheoremaLanguage", with one important
	difference: Functions "addNewKB" and "joinNewKB" are used rather than "appendKB" and "joinKB".
	Note that the names of the inference rules are in most cases the ones of the rules in "BasicTheoremaLanguage",
	because the proof texts are identical anyway.
*)

inferenceRule[ CGB2notGoal] = 
PRFSIT$[ g:FML$[ _, Not$TM[ a_], __], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {opp},
		opp = makeAssumptionFML[ formula -> a];
		makeANDNODE[ makePRFINFO[ name -> notGoal, used -> g], 
			toBeProved[ goal -> makeFML[ formula -> False, label -> "\[UpTee]"], kb -> addNewKB[ k, opp, Mode -> Prepend, Equality -> Weak], rest]
		]
	]
]

inferenceRule[ CGB2andKB] = 
PRFSIT$[ g_, k:{___FML$, FML$[ _, And$TM[ __], __], ___FML$}, id_, rest___?OptionQ] :> performProofStep[
	Module[ {conj = And$TM[], newKB, conjunctions},
		conjunctions = Cases[ k, FML$[ _, And$TM[ __], __]];
		newKB = DeleteCases[ k, FML$[ _, And$TM[ __], __]];
		Scan[Function[ f,
				conj = Join[ conj, MapIndexed[ makeAssumptionFML[ formula -> #1, label -> label@f <> "." <> ToString[ #2[[1]]]]&, formula@f]]
			],
			conjunctions
		];
		makeANDNODE[ makePRFINFO[ name -> CGB2andKB, used -> conjunctions], 
			toBeProved[ goal -> g, kb -> joinNewKB[ newKB, Apply[ List, conj], Mode -> Prepend, Equality -> Weak], rest]
		]
	]
]

inferenceRule[ CGB2andGoal] = 
PRFSIT$[ g:FML$[ _, And$TM[ c__], lab_, ___] /; FreeQ[ {c}, _META$], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {conjG, conjA, knew = k, tmp},
		conjG = MapIndexed[ makeGoalFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]]]&, {c}];
		conjA = MapIndexed[ makeAssumptionFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]]]&, {c}];
		makeANDNODE[ makePRFINFO[ name -> CGB2andGoal, used -> g], 
			MapThread[ (tmp = toBeProved[ goal -> #1, kb -> knew, rest]; knew = addNewKB[ knew, #2, Mode -> Append, Equality -> Weak]; tmp)&, {conjG, conjA}]
		]
	]
]

inferenceRule[ CGB2orGoal] = 
PRFSIT$[ g:FML$[ _, Or$TM[ a__, b_], lab_, ___] /; FreeQ[ {a, b}, _META$], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {negAssum, newG},
		negAssum = MapIndexed[ makeAssumptionFML[ formula -> sinkNegation[ Not$TM[#1]], label -> lab <> "." <> ToString[ #2[[1]]]]&, {a}];
		newG = makeGoalFML[ formula -> b];
		makeANDNODE[ makePRFINFO[ name -> orGoal, used -> g], 
			toBeProved[ goal -> newG, kb -> joinNewKB[ k, negAssum, Mode -> Prepend, Equality -> Weak], rest]
		]
	]
]

inferenceRule[ CGB2orGoalInteractive] = 
PRFSIT$[ g:FML$[ _, Or$TM[ a_, b__], lab_, ___] /; FreeQ[ {a, b}, _META$], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {newGoals, negAssm, newG, sel},
		newGoals = MapIndexed[ makeGoalFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]]]&, {a, b}];
		sel = selectGoalDialog[ newGoals, k, "goalHeader", "!goal"];
		NotebookClose[ $TMAproofNotebook];
		If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ newGoals],
			newG = newGoals[[sel]];
			negAssm = DeleteCases[ MapIndexed[ If[ #2 === {sel},
											$Failed,
											makeAssumptionFML[ formula -> sinkNegation[ Not$TM[ #1]], label -> lab <> "." <> ToString[ #2[[1]]]]
										]&, {a, b}
									],
					  				$Failed
					  ];
			makeANDNODE[ makePRFINFO[ name -> orGoal, used -> g, generated -> {Append[ negAssm, newG]}], 
				toBeProved[ goal -> newG, kb -> joinNewKB[ k, negAssm, Mode -> Prepend, Equality -> Weak], rest]
			]
			,
			$Failed
		]
	]
]

inferenceRule[ CGB2orKB] = 
PRFSIT$[ g_, {pre___, k:FML$[ _, Or$TM[ c__], lab_, ___], post___}, id_, rest___?OptionQ] :> performProofStep[
	Module[ {caseAssum},
		caseAssum = MapIndexed[ makeAssumptionFML[ formula -> #1, label -> lab <> "." <> ToString[ #2[[1]]]]&, {c}];
		makeANDNODE[ makePRFINFO[ name -> CGB2orKB, used -> k], 
			Map[ Block[ {$rewriteRules = {}}, toBeProved[ goal -> g, kb -> addNewKB[{pre, post}, #, Mode -> Prepend, Equality -> Weak], rest]]&, caseAssum]
		]
	]
]

inferenceRule[ CGB2implGoalDirect] = 
PRFSIT$[ g:FML$[ _, Implies$TM[ P_, Q_], __], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {left, right},
		left = makeAssumptionFML[ formula -> P];
		right = makeGoalFML[ formula -> Q];
		makeANDNODE[ makePRFINFO[ name -> implGoalDirect, used -> g], 
			toBeProved[ goal -> right, kb -> addNewKB[ k, left, Mode -> Prepend, Equality -> Weak], rest]]
	]
]

inferenceRule[ CGB2equivKB] = 
PRFSIT$[ g_FML$, k:{___FML$, FML$[ _, Iff$TM[ _, _], __], ___FML$}, id_, rest___?OptionQ] :> performProofStep[
	Module[ {impl = {}, newKB, equivalences, lhs, rhs},
		equivalences = Cases[ k, FML$[ _, Iff$TM[ _, _], __]];
		newKB = DeleteCases[ k, FML$[ _, Iff$TM[ _, _], __]];
		Scan[Function[ fml,
				lhs = fml[[2, 1]];
				rhs = fml[[2, 2]];
				AppendTo[ impl, {makeAssumptionFML[ formula -> Implies$TM[ lhs, rhs], label -> label@fml <> ".\[DoubleRightArrow]"],
								makeAssumptionFML[ formula -> Implies$TM[ rhs, lhs], label -> label@fml <> ".\[DoubleLeftArrow]"]}];
			],
			equivalences
		];
		makeANDNODE[ makePRFINFO[ name -> CGB2equivKB, used -> equivalences, generated -> impl], 
			toBeProved[ goal -> g, kb -> joinNewKB[ newKB, Flatten[ impl, 1], Mode -> Prepend, Equality -> Weak], rest]
		]
	]
]


(* ::Section:: *)
(* Miscellaneous *)


inferenceRule[ CGB2exchangeGoalKBInteractive] = 
PRFSIT$[ g_, k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {u = {g}, gen, negGoal, negFml, fml, sel, knew = If[ MemberQ[ k, FML$[ _, True|Not$TM[ False], ___]], k, Append[ k, makeFML[ formula -> True, label -> "True", key -> {"$True", "$None"}]]]},
		sel = selectAssumptionDialog[ g, knew, "exchangeHeader", "!exchange"];
		NotebookClose[ $TMAproofNotebook];
		If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ knew],
			negGoal = makeAssumptionFML[ formula -> sinkNegation[ Not$TM[ formula@g]]];
			gen = {negGoal};
			fml = knew[[sel]];
			negFml = makeGoalFML[ formula -> sinkNegation[ Not$TM[ formula@fml]]];
			If[ id@fml === "$True" && sel === Length[ knew],
				knew = k,
            	AppendTo[ u, fml];
            	AppendTo[ gen, negFml];
            	knew = Delete[ k, sel]
			];
			makeANDNODE[ makePRFINFO[ name -> CGB2exchangeGoalKBInteractive, used -> u, generated -> gen], 
                	toBeProved[ goal -> negFml,
                				kb -> addNewKB[ knew, negGoal, Mode -> Prepend, Equality -> Weak],
                				rest]
            ],
			$Failed
		]
	]
]

inferenceRule[ CGB2expandNegationKB] = 
PRFSIT$[ g_, k:{pre___FML$, fml:FML$[ _, f:Not$TM[ _Not$TM|_And$TM|_Or$TM|_Implies$TM|_Iff$TM|_Exists$TM|_Forall$TM], __], post___FML$}, _, rest___?OptionQ] :> performProofStep[
	Module[ {new = makeAssumptionFML[ formula -> sinkNegation[ f]]},
		makeANDNODE[ makePRFINFO[ name -> CGB2expandNegationKB, used -> fml, generated -> new],
        		toBeProved[ goal -> g,
                			kb -> addNewKB[ {pre, post}, new, Mode -> Prepend, Equality -> Weak],
                			rest]
		]
	]
]

inferenceRule[ CGB2expandNegationGoal] = 
PRFSIT$[ g:FML$[ _, f:Not$TM[ _Not$TM|_And$TM|_Or$TM|_Implies$TM|_Iff$TM|_Exists$TM|_Forall$TM], __], k_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {new = makeGoalFML[ formula -> sinkNegation[ f]]},
		makeANDNODE[ makePRFINFO[ name -> CGB2expandNegationGoal, used -> g, generated -> new],
        		toBeProved[ goal -> new, kb -> k, rest]
		]
	]
]


(* ::Section:: *)
(* Integers *)
	
inferenceRule[ CGB2integerIntervalKB] =
PRFSIT$[ g_FML$, KB:{KBa___, k:FML$[ _, DomainOperation$TM[ IntegerInterval$TM[ a:Except[ DirectedInfinity[-1]], b:Except[ DirectedInfinity[1]], True, True], Element$TM][ x_], __], KBb___}, _, rest___?OptionQ] :>
performProofStep[
   Module[{knew, gen},
	   If[ IntegerQ[ a] && IntegerQ[ b],
	    	If[ a <= b,
	      		gen = {makeAssumptionFML[ formula -> Apply[ Or$TM, Map[ Equal$TM[ x, Plus[a, #]]&, Range[ 0, b - a]]]], 
	         		   makeAssumptionFML[ formula -> DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Element$TM][ x]]
	      			  };
	      		makeANDNODE[ makePRFINFO[ name -> CGB2integerIntervalKB, used -> k, generated -> gen], 
	       			toBeProved[ goal -> g, kb -> joinNewKB[{KBa, KBb}, gen, Equality -> Weak, Mode -> Append], rest]
	          	],
		      	makeTERMINALNODE[ makePRFINFO[ name -> CGB2integerIntervalKB, used -> k], proved]
		  	],
	     	gen = {
		     		makeAssumptionFML[ formula -> DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Element$TM][ x] ], 
		       		makeAssumptionFML[ formula -> DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], LessEqual$TM][ a, x] ], 
		       		makeAssumptionFML[ formula -> DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], LessEqual$TM][ x, b] ]
	       		  };
	     	knew = joinNewKB[ {KBa, KBb}, gen, Equality -> Weak, Mode -> Append];
	  		makeANDNODE[ makePRFINFO[ name -> CGB2integerIntervalKB, used -> k, generated -> gen], 
	    				toBeProved[ goal -> g, kb -> knew, rest]
	    				]
	    ]
	]
]

inferenceRule[ CGB2integerIntervalGoal] =
PRFSIT$[ g:FML$[_, DomainOperation$TM[ IntegerInterval$TM[ a:Except[ DirectedInfinity[-1]], b:Except[ DirectedInfinity[1]], True, True], Element$TM][ x_], __], KB_List, _, rest___?OptionQ] :>
performProofStep[
	Module[ {gen},
		gen = makeGoalFML[ formula -> And$TM[
					DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Element$TM][ x],
					DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], LessEqual$TM][ a, x], 
					DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], LessEqual$TM][ x, b] 
				]
			];
		makeANDNODE[ makePRFINFO[ name -> CGB2integerIntervalGoal, used -> g, generated -> gen],
				toBeProved[ goal -> gen, kb -> KB, rest]
			]
	]
]


(* ::Section:: *)
(* Tuples *)

inferenceRule[ CGB2subscriptTupleOf] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {constants, const, tuple, tmpKB, newK, cond, R, t, phi, var, abf, RC, newG, gen, form, locC},
		constants = Cases[ KB, FML$[ _, DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], _, _], LessEqual$TM][ c_FIX$, a_], __] :> {c, a}];
		locC = getOptionalComponent[ this, "constants"];
		Catch[
			Scan[ Function[ c,
					const = First[ c];
					tuple = DeleteDuplicates[ Cases[ {g, KB}, Subscript$TM[ tup_TupleOf$TM, const] :> tup, Infinity, Heads -> True]];
					If[ Length[ tuple] === 1,
						tuple = First[ tuple];
						{R, t, phi} = First[ Cases[ {tuple}, TupleOf$TM[ rng_RNG$, p_, term_] :> {rng, term, p}]];
						var = specifiedVariables[ R];
						If[ Length[ var] === 1,
							var = First[ var];
							If[ FreeQ[ g /. Subscript$TM[ tuple, const] :> Null, const],
								newK = DeleteCases[ KB, FML$[ _, DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], _, _], LessEqual$TM][ 1, const], __]];
								If[ Length[ newK] < Length[ KB],
									tmpKB = DeleteCases[ newK, FML$[ _, DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], _, _], Element$TM][ const], __]];
									tmpKB = DeleteCases[ tmpKB, FML$[ _, DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], _, _], LessEqual$TM][ const, _], __]];
									If[ Length[ tmpKB] < Length[ newK] && FreeQ[ tmpKB /. Subscript$TM[ tuple, const] :> Null, const],
										cond = makeConditionFML[ formula -> And$TM[ isTuple$TM[ tuple], Equal$TM[ Last[c], BracketingBar$TM[ tuple]]]];
										abf = newABF[ var, {g, KB}];
										t = substituteFree[ t, {VAR$[ var] -> abf}];
										phi = makeAssumptionFML[ formula -> substituteFree[ phi, {VAR$[ var] -> abf}]];
										RC = makeAssumptionFML[ formula -> First[ myRngToCondition[ R]] /. VAR$[ var] -> abf];
										newG = makeGoalFML[ formula -> formula@g /. Subscript$TM[ tuple, const] :> t];
										newK = {};
										gen = {};
										Scan[ Function[ f,
												form = formula@f /. Subscript$TM[ tuple, const] :> t;
												If[ form === formula@f,
													newK = addNewKB[ newK, f, Mode -> Append],
													form = makeAssumptionFML[ formula -> form];
													newK = addNewKB[ newK, form, Mode -> Append];
													AppendTo[ gen, form];
												]
											],
											tmpKB
										];
										Throw[
											makeANDNODE[ makePRFINFO[ name -> CGB2subscriptTupleOf, used -> {{tuple, const}}, generated -> {{abf, cond, t, phi, RC, newG, gen}}],
														{toBeProved[ goal -> cond, kb -> newK, rest],
														toBeProved[ goal -> newG, kb -> joinNewKB[ newK, {cond, phi, RC}], "constants" -> Prepend[ locC, RNG$[ STEPRNG$[ abf, 1, BracketingBar$TM[ tuple], 1]]], rest]}
											]
										]
									]
								]
							]
						]
					]
				],
				constants
			];
			$Failed
		]
	]
]

inferenceRule[ CGB2elementTupleOf] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {usedTuples, list, S, j, t, i, u, form, cond, newG, newK = {}, gen = {}},
		usedTuples = getOptionalComponent[ this, "elementTupleOf"];
		list = Cases[ {g, KB},
					  (Subscript$TM[ tuple:TupleOf$TM[ RNG$[ STEPRNG$[ var_, 1, upper_, 1]], True, term_], index_] /;
									 freeVariables[ tuple] === {} && freeVariables[ index] === {} && !MemberQ[ usedTuples, {tuple, index}]) :>
							{tuple, index, var, upper, term},
					  Infinity, 1, Heads -> True
				];
		If[ list === {},
			$Failed,
			{S, j, i, u, t} = First[ list];
			t = substituteFree[ t, {i -> j}];
			form = formula@g /. Subscript$TM[ S, j] :> t;
			If[ form === formula@g,
				newG = g,
				newG = makeGoalFML[ formula -> form]
			];
			Scan[ Function[ f,
					form = formula@f /. Subscript$TM[ S, j] :> t;
					If[ form === formula@f,
						newK = addNewKB[ newK, f, Mode -> Append],
						form = makeAssumptionFML[ formula -> form];
						AppendTo[ gen, form];
						newK = addNewKB[ newK, form, Mode -> Append]
					]
				],
				KB
			];
			If[ MemberQ[ usedTuples, {S, j, "checked"}],
				makeANDNODE[ makePRFINFO[ name -> CGB2elementTupleOf, used -> {{S, j}}, generated -> If[ newG === g, {gen, {t}}, {{newG}, gen, {t}}]],
							toBeProved[ goal -> newG, kb -> newK, rest]
				],
				cond = makeConditionFML[ formula -> And$TM[ isTuple$TM[ S], DomainOperation$TM[ IntegerInterval$TM[ 1, u, True, True], Element$TM][ j]]];
				makeANDNODE[ makePRFINFO[ name -> CGB2elementTupleOf, used -> {{S, j}}, generated -> {{cond, newG}, gen, {t}}],
							{Block[ {$rewriteRules = {}}, toBeProved[ goal -> cond, kb -> KB, "elementTupleOf" -> Prepend[ usedTuples, {S, j}], rest]],
							toBeProved[ goal -> newG, kb -> addNewKB[ newK, cond], "elementTupleOf" -> Prepend[ usedTuples, {S, j, "checked"}], rest]}
				]
			]
		]
	]
]

inferenceRule[ CGB2lengthTupleOf] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {usedTuples, list, S, u, form, newG, newK = {}, gen = {}, cond},
		usedTuples = getOptionalComponent[ this, "lengthTupleOf"];
		list = Cases[ {g, KB},
					  (BracketingBar$TM[ tuple:TupleOf$TM[ RNG$[ STEPRNG$[ _, 1, upper_, 1]], True, _]] /;
									 freeVariables[ tuple] === {} && !MemberQ[ usedTuples, tuple]) :>
							{tuple, upper},
					  Infinity, 1, Heads -> True
				];
		If[ list === {},
			$Failed,
			{S, u} = First[ list];
			form = formula@g /. BracketingBar$TM[ S] :> u;
			If[ form === formula@g,
				newG = g,
				newG = makeGoalFML[ formula -> form]
			];
			Scan[ Function[ f,
					form = formula@f /. BracketingBar$TM[ S] :> u;
					If[ form === formula@f,
						newK = addNewKB[ newK, f, Mode -> Append],
						form = makeAssumptionFML[ formula -> form];
						AppendTo[ gen, form];
						newK = addNewKB[ newK, form, Mode -> Append]
					]
				],
				KB
			];
			If[ MemberQ[ usedTuples, {S, "checked"}],
				makeANDNODE[ makePRFINFO[ name -> CGB2lengthTupleOf, used -> {{S}}, generated -> If[ newG === g, {gen, {u}}, {{newG}, gen, {u}}]],
							toBeProved[ goal -> newG, kb -> newK, rest]
				],
				cond = makeConditionFML[ formula -> isTuple$TM[ S]];
				makeANDNODE[ makePRFINFO[ name -> CGB2lengthTupleOf, used -> {{S}}, generated -> {{cond, newG}, gen, {u}}],
							{Block[ {$rewriteRules = {}}, toBeProved[ goal -> cond, kb -> KB, "lengthTupleOf" -> Prepend[ usedTuples, S], rest]],
							toBeProved[ goal -> newG, kb -> addNewKB[ newK, cond], "lengthTupleOf" -> Prepend[ usedTuples, {S, "checked"}], rest]}
				]
			]
		]
	]
]

inferenceRule[ CGB2nonemptyTupleOf] =
PRFSIT$[ g:FML$[ _, DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], LessEqual$TM][ 1, BracketingBar$TM[ T:TupleOf$TM[ r_RNG$, phi_, _]]], __], KB_List, _, rest___?OptionQ] :>
performProofStep[
	Module[ {gen},
		gen = makeGoalFML[ formula -> And$TM[ isTuple$TM[ T], Exists$TM[ r, True, phi]]];
		makeANDNODE[ makePRFINFO[ name->CGB2nonemptyTupleOf, used -> g, generated -> gen],
					toBeProved[ goal -> gen, kb -> KB, rest]
		]
	]
]

inferenceRule[ CGB2isTupleTupleOf] =
PRFSIT$[ g:FML$[ _, isTuple$TM[ T:TupleOf$TM[ RNG$[ STEPRNG$[ _VAR$, a_, b_, 1]], _, _]], __], KB_List, _, rest___?OptionQ] :>
performProofStep[
	Module[ {gen},
		gen = makeGoalFML[ formula -> And$TM[ DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Element$TM][ a],
											DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Element$TM][ b]]
			];
		makeANDNODE[ makePRFINFO[ name -> CGB2isTupleTupleOf, used -> g, generated -> gen],
					toBeProved[ goal -> gen, kb -> KB, rest]
		]
	]
]

inferenceRule[ CGB2appendElement] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {usedTuples, list, A, x, t, form, rules, newG = {Null, Null}, newK = {{}, {}}, gen = {{}, {}}, cond, isTuple, ind, i, index, tbp = {}},
		usedTuples = getOptionalComponent[ this, "appendElement"];
		list = Cases[ {g, KB},
					  (Subscript$TM[ appendElem$TM[ tuple:Except[ _Tuple$TM], elem_], sub_] /;
									 freeVariables[ tuple] === freeVariables[ elem] === freeVariables[ sub] === {} && !MemberQ[ usedTuples, {tuple, elem, sub}]) :>
							{tuple, elem, sub},
					  Infinity, 1, Heads -> True
				];
		If[ list === {},
			$Failed,
			{A, x, t} = First[ list];
			rules = {Subscript$TM[ appendElem$TM[ A, x], t] :> Subscript$TM[ A, t],
					Subscript$TM[ appendElem$TM[ A, x], t] :> x};
			ind = Cases[ usedTuples, {A, x, t, a_Integer} :> a, {1}, 1];
			If[ ind === {},
				cond = makeGoalFML[ formula -> And$TM[ isTuple$TM[ A],
								DomainOperation$TM[ IntegerInterval$TM[ 1, BracketingBar$TM[ appendElem$TM[ A, x]], True, True], Element$TM][ t]]];
				isTuple = makeAssumptionFML[ formula -> isTuple$TM[ A]];
				newK = {{isTuple,
							makeAssumptionFML[ formula -> DomainOperation$TM[ IntegerInterval$TM[ 1, BracketingBar$TM[ A], True, True], Element$TM][ t]]},
					 	{isTuple,
					 		makeAssumptionFML[ formula -> DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Element$TM][ t]],
					 		makeAssumptionFML[ formula -> Equal$TM[ t, DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Plus$TM][ BracketingBar$TM[ A], 1]]]}
					 	};
				AppendTo[ gen, Prepend[ newK, cond]];
				ind = {1, 2};
				tbp = {Block[ {$rewriteRules = {}}, toBeProved[ goal -> cond, kb -> KB, "appendElement" -> Prepend[ usedTuples, {A, x, t}], rest]]};
			];
			For[ i = 1, i <= Length[ ind], i++,
				index =  ind[[i]];
				form = formula@g /. rules[[index]];
				If[ form === formula@g,
					newG[[index]] = g,
					newG[[index]] = makeGoalFML[ formula -> form];
					gen[[index]] = {{newG[[index]]}}
				];
				Scan[ Function[ f,
						form = formula@f /. rules[[index]];
						If[ form === formula@f,
							AppendTo[ newK[[index]], f],
							form = makeAssumptionFML[ formula -> form];
							AppendTo[ gen[[index]], form];
							AppendTo[ newK[[index]], form]
						]
					],
					KB
				];
				(* We have to create the new knowledge base in such complicated way because we have to make sure that each branch gets the right rewrite rules. *)
				AppendTo[ tbp, Block[ {$rewriteRules = {}}, toBeProved[ goal -> newG[[index]], kb -> joinNewKB[ {}, newK[[index]]], rest, "appendElement" -> Prepend[ usedTuples, {A, x, t, index}]]]]
			];
			makeANDNODE[ makePRFINFO[ name -> CGB2appendElement, used -> {{A, x, t}}, generated -> gen], tbp]
		]
	]
]

inferenceRule[ CGB2appendTupleOf] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {usedTuples, list, tupleof, A, x, i, t, phi, form, expr, rules, newG = {Null, Null}, newK = {{}, {}}, gen = {{}, {}}, cond, tbp, fphi, j},
		usedTuples = getOptionalComponent[ this, "appendTupleOf"];
		list = Cases[ {g, KB},
					  (to:(TupleOf$TM[ RNG$[ STEPRNG$[ var_, 1, BracketingBar$TM[ app:(appendElem$TM[ tuple_, elem_])], 1]], c_, term_]) /;
								freeVariables[ tuple] === freeVariables[ elem] === freeVariables[ c /. Subscript$TM[ app, var] -> Null] === freeVariables[ term /. Subscript$TM[ app, var] -> Null] === {} && !MemberQ[ usedTuples, to]) :>
							{to, tuple, elem, var, c, term},
					  Infinity, 1, Heads -> True
				];
		If[ list === {},
			$Failed,
			
			{tupleof, A, x, i, phi, t} = First[ list];
			PrependTo[ usedTuples, tupleof];
			cond = makeConditionFML[ formula -> isTuple$TM[ A]];
			tbp = {Block[ {$rewriteRules = {}}, toBeProved[ goal -> cond, kb -> KB, rest, "appendTupleOf" -> usedTuples]]};
			
			fphi = {makeAssumptionFML[ formula -> (phi /. Subscript$TM[ appendElem$TM[ A, x], i] :> x)],
					makeAssumptionFML[ formula -> Not$TM[ (phi /. Subscript$TM[ appendElem$TM[ A, x], i] :> x)]]};
			AppendTo[ gen, {cond, fphi}];
			
			expr = TupleOf$TM[ RNG$[ STEPRNG$[ i, 1, BracketingBar$TM[ A], 1]], phi, t]  /. Subscript$TM[ appendElem$TM[ A, x], i] :> Subscript$TM[ A, i];
			rules = {tupleof :> appendElem$TM[ expr, t /. Subscript$TM[ appendElem$TM[ A, x], i] :> x], tupleof :> expr};
			
			For[ j = 1, j <= 2, j++,
				form = formula@g /. rules[[j]];
				If[ form === formula@g,
					newG[[j]] = g,
					newG[[j]] = makeGoalFML[ formula -> form];
					gen[[j]] = {{newG[[j]]}}
				];
				Scan[ Function[ f,
						form = formula@f /. rules[[j]];
						If[ form === formula@f,
							AppendTo[ newK[[j]], f],
							form = makeAssumptionFML[ formula -> form];
							AppendTo[ gen[[j]], form];
							AppendTo[ newK[[j]], form]
						]
					],
					KB
				];
				AppendTo[ tbp, Block[ {$rewriteRules = {}}, toBeProved[ goal -> newG[[j]], kb -> joinNewKB[ {cond, fphi[[j]]}, newK[[j]]], rest, "appendElement" -> usedTuples]]]
			];
			makeANDNODE[ makePRFINFO[ name -> CGB2appendTupleOf, used -> {{tupleof, A, rules[[1, 2]], rules[[2, 2]]}}, generated -> gen], tbp]
		]
	]
]


(* ::Section:: *)
(* Ordering *)

inferenceRule[ CGB2contradictionOrdering] = 
PRFSIT$[ g_, KB:{pre___, k:FML$[ _, DomainOperation$TM[ dom_, Less$TM][ x_, x_], __], post___},
		_, rest___?OptionQ] :> performProofStep[
	Module[ {newG},
		newG = makeGoalFML[ formula -> DomainOperation$TM[ dom, Element$TM][ x]];
		makeANDNODE[ makePRFINFO[ name -> CGB2contradictionOrdering, used -> {k}, generated -> {newG}],
    			toBeProved[ goal -> newG, kb -> {pre, post}, rest]
		]
	]
] /; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTotallyOrdered$TM|Theorema`Knowledge`isTOM$TM)[ dom], __]]

inferenceRule[ CGB2notLessEqual] =
(PRFSIT$[ g:FML$[ _, form:Not$TM[ DomainOperation$TM[ dom_, operation:(Less$TM|LessEqual$TM)][ x_, y_]], __], KB_List, _, rest___?OptionQ])|(PRFSIT$[ g_FML$, KB:{___, FML$[ _, form:Not$TM[ DomainOperation$TM[ dom_, operation:(Less$TM|LessEqual$TM)][ x_, y_]], __], ___}, _, rest___?OptionQ]) :>
performProofStep[
	Module[ {gen, newg, newkb, newf, newg2, newkb2 = {}, newoperation},
		(*deciding operation depending for replacing occurences*)
		newoperation = Switch[ operation,
								LessEqual$TM,
								Less$TM,
								Less$TM,
								LessEqual$TM
						];
		newkb = DeleteCases[ KB, FML$[ _, Not$TM[ DomainOperation$TM[ dom, operation][ x, y]], __]];
		newg = makeConditionFML[ formula -> And$TM[ DomainOperation$TM[ dom, Element$TM][ x], DomainOperation$TM[ dom, Element$TM][ y]]];
		gen = {DomainOperation$TM[ dom, newoperation][ y, x], newg};
		newg2 = formula@g /. Not$TM[ DomainOperation$TM[ dom, operation][ x, y]] :> DomainOperation$TM[ dom, newoperation][ y, x];
		If[ newg2 === formula@g,
			newg2 = g,
			newg2 = makeGoalFML[ formula -> newg2];
		];
		AppendTo[ gen, newg2];
		
		newkb2 = {newg};
		Scan[ Function[ f,
				newf = formula@f /. Not$TM[ DomainOperation$TM[ dom, operation][ x, y]] :> DomainOperation$TM[ dom, newoperation][ y, x];
				If[ newf === formula@f,
					newkb2 = addNewKB[ newkb2, f, Mode -> Append],
					newf = makeAssumptionFML[ formula -> newf];
					newkb2 = addNewKB[ newkb2, newf, Mode -> Append];
					AppendTo[ gen, newf]
				]
			],
			KB
		];
		
		makeANDNODE[ makePRFINFO[ name -> CGB2notLessEqual, used -> {{form}}, generated -> {gen}],
					{Block[ {$rewriteRules = {}}, toBeProved[ goal -> newg, kb -> newkb, rest]],
					toBeProved[ goal -> newg2, kb -> newkb2, rest]}
		]
	]
] /; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTotallyOrdered$TM|Theorema`Knowledge`isTOM$TM)[ dom], __]]

inferenceRule[ CGB2reflexiveClosure] =
PRFSIT$[ g_FML$, KB:{ KBa___, k:FML$[ _, DomainOperation$TM[ dom_, Less$TM][ x_, y_],__], KBb___}, _, rest___?OptionQ] :>
performProofStep[
	Module[ {gen, newg},
		newg = makeGoalFML[ formula -> And$TM[ DomainOperation$TM[ dom, Element$TM][ x], DomainOperation$TM[ dom, Element$TM][ y]]];
		gen = {makeAssumptionFML[ formula -> DomainOperation$TM[ dom, Element$TM][ x]],
			makeAssumptionFML[ formula -> DomainOperation$TM[ dom, Element$TM][ y]],
			makeAssumptionFML[ formula -> DomainOperation$TM[ dom, LessEqual$TM][ x, y]]};
		makeANDNODE[ makePRFINFO[ name -> CGB2reflexiveClosure, used -> {{k, newg, dom, x, y, g}}, generated -> gen],
					{Block[ {$rewriteRules = {}}, toBeProved[ goal -> newg, kb -> {KBa, KBb}, rest]],
					toBeProved[ goal -> g, kb -> joinNewKB[ {KBa, k, KBb}, gen, Equality -> Weak, Mode -> Append], rest]}
		]
	]
] /; (MemberQ[ {KBa, KBb}, FML$[ _, (Theorema`Knowledge`isTotallyOrdered$TM|Theorema`Knowledge`isTOM$TM)[ dom], __]] &&
	  !MemberQ[ {KBa, KBb}, FML$[ _, DomainOperation$TM[ dom, LessEqual$TM][ x, y], __]])

inferenceRule[ CGB2transitivityKB] = 
this:PRFSIT$[ g_, KB:({___, FML$[ _, DomainOperation$TM[ dom_, Less$TM|LessEqual$TM][ x_, y_], __], ___, FML$[ _, DomainOperation$TM[ dom_, Less$TM|LessEqual$TM][ y_, z_], __], ___} |
				{___, FML$[ _, DomainOperation$TM[ dom_, Less$TM|LessEqual$TM][ y_, z_], __], ___, FML$[ _, DomainOperation$TM[ dom_, Less$TM|LessEqual$TM][ x_, y_], __], ___}),
		_, rest___?OptionQ] :> performProofStep[
	Module[ {newF, newG, k1, k2, d, a, b, c, r1, r2, usedFormulae},
		usedFormulae = getOptionalComponent[ this, "transitivityKB"];
		(* we want a pair of suitable formulae that we have not yet used for inferring a new formula by transitivity *)
		{k1, k2, d, a, b, c, r1, r2} = getFirstUnusedTransitivityFormulae[ KB, usedFormulae];
		If[ k1 =!= $Failed && k2 =!= $Failed,
			PrependTo[ usedFormulae, {key@k1, key@k2}];
			If[ a === c && SameQ[ r1, r2, LessEqual$TM],
				newG = makeGoalFML[ formula -> And$TM[ DomainOperation$TM[ d, Element$TM][ a], DomainOperation$TM[ d, Element$TM][ b]]];
				newF = makeAssumptionFML[ formula -> Equal$TM[ a, b]],
				
				newG = makeGoalFML[ formula -> DomainOperation$TM[ d, Element$TM][ b]];
				newF = makeAssumptionFML[ formula -> DomainOperation$TM[ d, If[ SameQ[ r1, r2, LessEqual$TM], LessEqual$TM, Less$TM]][ a, c]]
			];
			makeANDNODE[ makePRFINFO[ name -> CGB2transitivityKB, used -> {k1, k2}, generated -> {newG, newF}], 
        			{Block[ {$rewriteRules = {}}, toBeProved[ goal -> newG, kb -> KB, "transitivityKB" -> usedFormulae, rest]],
        			toBeProved[ goal -> g, kb -> addNewKB[ KB, newF, Equality -> Strong, Mode -> Append], "transitivityKB" -> usedFormulae, rest]}
			],
			$Failed
		]
	]
] /; (x =!= y && y =!= z)

inferenceRule[ CGB2transitivityGoal] =
PRFSIT$[ g:FML$[ _, DomainOperation$TM[ dom_, gRel:(Less$TM|LessEqual$TM)][ x_, y_], __],
		 KB:{___, k:FML$[ _, (DomainOperation$TM[ dom_, kRel:(Less$TM|LessEqual$TM)][ x_, z_])|(DomainOperation$TM[ dom_, kRel:(Less$TM|LessEqual$TM)][ z_, y_]), __], ___},
		 _, rest___?OptionQ] :>
performProofStep[
	Module[ {newG, newRel},
		newRel = Switch[ gRel,
					LessEqual$TM,
					LessEqual$TM,
					_,
					Switch[ kRel,
						LessEqual$TM,
						Less$TM,
						_,
						LessEqual$TM	
					]
				];
		newG = Switch[ k[[2, 1]],
					x,
					makeGoalFML[
						formula -> And$TM[ DomainOperation$TM[ dom, Element$TM][ x], DomainOperation$TM[ dom, newRel][ z, y]]
					],
					_,
					makeGoalFML[
						formula -> And$TM[ DomainOperation$TM[ dom, Element$TM][ y], DomainOperation$TM[ dom, newRel][ x, z]]
					]
				];
		makeANDNODE[ makePRFINFO[ name -> CGB2transitivityGoal, used -> {g, k}, generated -> {newG}],
					toBeProved[ goal -> newG, kb -> KB, rest]
		]
	]
] /; (x =!= y && MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTotallyOrdered$TM|Theorema`Knowledge`isTOM$TM)[ dom], __]])

(*
	Remark: For some strange reason, Mathematica gives warnings if we use a more complicated pattern involving
		PatternSequence in order to get rid of the second condition at the end of the function body.
*)
inferenceRule[ CGB2transitivityGoalAddition] = 
this:PRFSIT$[ g:FML$[ _, DomainOperation$TM[ dom_, rel1:(Less$TM|LessEqual$TM)][ lhs_, rhs_], __], KB_List, _, rest___?OptionQ] :>
	Module[ {left, right, newG, cond, condF, candidates, out = $Failed, form, rel2, a, b, atomicArgs, atomicArgs2, best = Null, bestLen = 0},
		Block[ {$rewriteRules = {}, $generated = {}},
			candidates = Cases[ KB, f:(FML$[ _, DomainOperation$TM[ dom, r2:(Less$TM|LessEqual$TM)][ x_, y_], __]) /; (x =!= y) :> {f, r2, {x, y}}];
			If[ candidates =!= {},
				If[ left,
					atomicArgs = getAllAtomicArguments[ lhs, dom, True];
					Scan[(
						{a, b} = Last[ #];
						atomicArgs2 = getAllAtomicArguments[ a, dom, True];
						If[ SublistQ[ atomicArgs2, atomicArgs] && Length[ atomicArgs2] > bestLen,
							bestLen = Length[ atomicArgs2];
							best = #
						]
						)&,
						candidates
					];
					If[ best =!= Null,
						{form, rel2, {a, b}} = best;
						
						(* collect type conditions of all elements in the lhs and in the KB relation that is used *)
						atomicArgs2 = getAllAtomicArguments[ a, dom, True];
						cond = Apply[ And$TM, Map[ DomainOperation$TM[ dom, Element$TM], Union[ atomicArgs, atomicArgs2, getAllAtomicArguments[ b, dom, True]]]];
						condF = makeConditionFML[ formula -> cond];
						
						(* replace arguments *)
						Scan[(
							atomicArgs = DeleteCases[ atomicArgs, #, {1}, Count[ atomicArgs2, #]]
							)&,
							DeleteDuplicates[ atomicArgs2]
						];
						rel2 = If[ rel1 === Less$TM && rel2 === LessEqual$TM, Less$TM, LessEqual$TM];
						newG = makeGoalFML[ formula -> DomainOperation$TM[ dom, rel2][ Fold[ DomainOperation$TM[ dom, Plus$TM], b, atomicArgs], rhs]];
						
						out = makeANDNODE[ makePRFINFO[ name -> CGB2transitivityGoalAddition, used -> {g, form}, generated -> {condF, newG}, "side" -> Left],
									{toBeProved[ goal -> condF, kb -> KB, rest],
									toBeProved[ goal -> newG, kb -> addNewKB[ KB, condF], rest]}];
					]
				];
				If[ out === $Failed,
					If[ right,
						atomicArgs = getAllAtomicArguments[ rhs, dom, True];
						Scan[(
							{a, b} = Last[ #];
							atomicArgs2 = getAllAtomicArguments[ b, dom, True];
							If[ SublistQ[ atomicArgs2, atomicArgs] && Length[ atomicArgs2] > bestLen,
								bestLen = Length[ atomicArgs2];
								best = #
							]
							)&,
							candidates
						];
						If[ best =!= Null,
							{form, rel2, {a, b}} = best;
							
							(* collect type conditions of all elements in the lhs and in the KB relation that is used *)
							atomicArgs2 = getAllAtomicArguments[ b, dom, True];
							cond = Apply[ And$TM, Map[ DomainOperation$TM[ dom, Element$TM], Union[ atomicArgs, getAllAtomicArguments[ a, dom, True], atomicArgs2]]];
							condF = makeConditionFML[ formula -> cond];
							
							(* replace arguments *)
							Scan[(
								atomicArgs = DeleteCases[ atomicArgs, #, {1}, Count[ atomicArgs2, #]]
								)&,
								DeleteDuplicates[ atomicArgs2]
							];
							rel2 = If[ rel1 === Less$TM && rel2 === LessEqual$TM, Less$TM, LessEqual$TM];
							newG = makeGoalFML[ formula -> DomainOperation$TM[ dom, rel2][ lhs, Fold[ DomainOperation$TM[ dom, Plus$TM], a, atomicArgs]]];
							
							out = makeANDNODE[ makePRFINFO[ name -> CGB2transitivityGoalAddition, used -> {g, form}, generated -> {condF, newG}, "side" -> Right],
										{toBeProved[ goal -> condF, kb -> KB, rest],
										toBeProved[ goal -> newG, kb -> addNewKB[ KB, condF], rest]}];
						]
					]
				]
			];
			out
		] /; (left = MatchQ[ lhs, DomainOperation$TM[ dom, Plus$TM][ __]]; right = MatchQ[ rhs, DomainOperation$TM[ dom, Plus$TM][ __]]; left || right)
	] /; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]]
	
(* Does the same as "transitivityGoalAddition", except that it only considers the left-hand-side of inequalities *)
inferenceRule[ CGB2transitivityGoalAdditionLeft] = 
this:PRFSIT$[ g:FML$[ _, DomainOperation$TM[ dom_, rel1:(Less$TM|LessEqual$TM)][ lhs:(DomainOperation$TM[ dom_, Plus$TM][ __]), rhs_], __], KB_List, _, rest___?OptionQ] :>
	Module[ {newG, cond, condF, candidates, out = $Failed, form, rel2, a, b, atomicArgs, atomicArgs2, best = Null, bestLen = 0},
		Block[ {$rewriteRules = {}, $generated = {}},
			candidates = Cases[ KB, f:(FML$[ _, DomainOperation$TM[ dom, r2:(Less$TM|LessEqual$TM)][ x_, y_], __]) /; (x =!= y) :> {f, r2, {x, y}}];
			If[ candidates =!= {},
				atomicArgs = getAllAtomicArguments[ lhs, dom, True];
				Scan[(
					{a, b} = Last[ #];
					atomicArgs2 = getAllAtomicArguments[ a, dom, True];
					If[ SublistQ[ atomicArgs2, atomicArgs] && Length[ atomicArgs2] > bestLen,
						bestLen = Length[ atomicArgs2];
						best = #
					]
					)&,
					candidates
				];
				If[ best =!= Null,
					{form, rel2, {a, b}} = best;
					
					(* collect type conditions of all elements in the lhs and in the KB relation that is used *)
					atomicArgs2 = getAllAtomicArguments[ a, dom, True];
					cond = Apply[ And$TM, Map[ DomainOperation$TM[ dom, Element$TM], Union[ atomicArgs, atomicArgs2, getAllAtomicArguments[ b, dom, True]]]];
					condF = makeConditionFML[ formula -> cond];
					
					(* replace arguments *)
					Scan[(
						atomicArgs = DeleteCases[ atomicArgs, #, {1}, Count[ atomicArgs2, #]]
						)&,
						DeleteDuplicates[ atomicArgs2]
					];
					rel2 = If[ rel1 === Less$TM && rel2 === LessEqual$TM, Less$TM, LessEqual$TM];
					newG = makeGoalFML[ formula -> DomainOperation$TM[ dom, rel2][ Fold[ DomainOperation$TM[ dom, Plus$TM], b, atomicArgs], rhs]];
					
					out = makeANDNODE[ makePRFINFO[ name -> CGB2transitivityGoalAdditionLeft, used -> {g, form}, generated -> {condF, newG}, "side" -> Left],
								{toBeProved[ goal -> condF, kb -> KB, rest],
								toBeProved[ goal -> newG, kb -> addNewKB[ KB, condF], rest]}];
				]
			];
			out
		]
	] /; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]]
	
	
(* ::Section:: *)
(* Addition *)

inferenceRule[ CGB2cancellationGoal] = 
this:PRFSIT$[ g:FML$[ _, (r:(Equal$TM|DomainOperation$TM[ d_, Less$TM|LessEqual$TM]))[ lhs_, rhs_], __], KB_List, _, rest___?OptionQ] :>
	Module[ {dom, lFlat, rFlat, lElems, allElems, dc, rc, newL, newR, newG, condF, cancellation = False},
		Block[ {$rewriteRules = {}, $generated = {}},
			lFlat = getAllAtomicArguments[ lhs, dom, True];
			rFlat = getAllAtomicArguments[ rhs, dom, True];
			lElems = DeleteDuplicates[ lFlat];
			allElems = Apply[ And$TM, Union[ lElems, rFlat]];
			Scan[(
				rc = Count[ rFlat, #];
				If[ rc > 0,
					dc = Min[Count[ lFlat, #], rc];
					lFlat = DeleteCases[ lFlat, #, {1}, dc];
					rFlat = DeleteCases[ rFlat, #, {1}, dc];
					cancellation = True
				]
				)&,
				lElems
			];
			If[ cancellation,
				condF = makeConditionFML[ formula -> Map[ (DomainOperation$TM[ dom, Element$TM][ #])&, allElems]];
				newL = If[ lFlat === {},
						DomainOperation$TM[ dom, 0],
						Fold[ DomainOperation$TM[ dom, Plus$TM], First[ lFlat], Rest[ lFlat]]
					];
				newR = If[ rFlat === {},
						DomainOperation$TM[ dom, 0],
						Fold[ DomainOperation$TM[ dom, Plus$TM], First[ rFlat], Rest[ rFlat]]
					];
				newG = makeGoalFML[ formula -> r[ newL, newR]];
				makeANDNODE[ makePRFINFO[ name -> CGB2cancellationGoal, used -> {g}, generated -> {condF, newG}],
								{toBeProved[ goal -> condF, kb -> KB, rest],
								toBeProved[ goal -> newG, kb -> addNewKB[ KB, condF, Equality -> Weak], rest]}
				],
				$Failed
			]
		] /; If[ r === Equal$TM,
				If[ (dom = Cases[ {lhs}, DomainOperation$TM[ d0_, Plus$TM][ __] -> d0]) =!= {},
					dom = First[ dom],
					dom = {}
				];
				If[ dom === {},
					If[ (dom = Cases[ {rhs}, DomainOperation$TM[ d0_, Plus$TM][ __] -> d0]) =!= {},
						dom = First[ dom];
						MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]],
						False
					],
					MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]]
				],
				dom = d;
				MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]]
			]
	]
	
inferenceRule[ CGB2cancellationKB] = 
this:PRFSIT$[ g_, KB:{___, FML$[ _, (Equal$TM[ DomainOperation$TM[ d_, Plus$TM][___], _])|
									(Equal$TM[ _, DomainOperation$TM[ d_, Plus$TM][___]])|
									(DomainOperation$TM[ d_, Less$TM|LessEqual$TM][ DomainOperation$TM[ d_, Plus$TM][___], _])|
									(DomainOperation$TM[ d_, Less$TM|LessEqual$TM][ _, DomainOperation$TM[ d_, Plus$TM][___]]),
								__], ___}, _, rest___?OptionQ] :>
performProofStep[
	Module[ {dom, lFlat, rFlat, lElems, allElems, dc, rc, newL, newR, newG, condF, newK, cancellation = False, r},
		Catch[
			Scan[ Function[ f,
					dom = Cases[ {f}, FML$[ _, (Equal$TM[ lhs:DomainOperation$TM[ d0_, Plus$TM][___], rhs_])|
												(Equal$TM[ lhs_, rhs:DomainOperation$TM[ d0_, Plus$TM][___]]),
										__] :> {Equal$TM, d0, lhs, rhs} /; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ d0], __]]
						];
					If[ dom === {},
						dom = Cases[ {f}, FML$[ _, ((r0:(DomainOperation$TM[ d0_, Less$TM|LessEqual$TM]))[ lhs:DomainOperation$TM[ d0_, Plus$TM][___], rhs_])|
													((r0:(DomainOperation$TM[ d0_, Less$TM|LessEqual$TM]))[ lhs_, rhs:DomainOperation$TM[ d0_, Plus$TM][___]]),
										__] :> {r0, d0, lhs, rhs} /; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ d0], __]]
							]
					];
					If[ dom =!= {},
						{{r, dom, lFlat, rFlat}} = dom;
						lFlat = getAllAtomicArguments[ lFlat, dom, True];
						rFlat = getAllAtomicArguments[ rFlat, dom, True];
						lElems = DeleteDuplicates[ lFlat];
						allElems = Apply[ And$TM, Union[ lElems, rFlat]];
						Scan[(
							rc = Count[ rFlat, #];
							If[ rc > 0,
								dc = Min[ Count[ lFlat, #], rc];
								lFlat = DeleteCases[ lFlat, #, {1}, dc];
								rFlat = DeleteCases[ rFlat, #, {1}, dc];
								cancellation = True
							]
							)&,
							lElems
						];
						If[ cancellation,
							condF = makeConditionFML[ formula -> Map[ (DomainOperation$TM[ dom, Element$TM][ #])&, allElems]];
							newL = If[ lFlat === {},
									DomainOperation$TM[ dom, 0],
									Fold[ DomainOperation$TM[ dom, Plus$TM], First[ lFlat], Rest[ lFlat]]
								];
							newR = If[ rFlat === {},
									DomainOperation$TM[ dom, 0],
									Fold[ DomainOperation$TM[ dom, Plus$TM], First[ rFlat], Rest[ rFlat]]
								];
							newG = makeGoalFML[ formula -> r[ newL, newR]];
							newK = DeleteCases[ KB, f];
							Throw[ makeANDNODE[ makePRFINFO[ name -> CGB2cancellationKB, used -> {g, f}, generated -> {condF, newG}],
											{Block[ {$rewriteRules = {}}, toBeProved[ goal -> condF, kb -> newK, rest]],
											toBeProved[ goal -> g, kb -> joinNewKB[ newK, {newG, condF}, Equality -> Weak], rest]}]
							]
						]
					]
				],
				KB
			];
			$Failed
		]
	]
] /; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ d], __]]


(* ::Section:: *)
(* ArgMin, ArgMax *)
	
inferenceRule[ CGB2ArgMin] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	abbreviateArgminmax[ this, ArgMin$TM, "ArgMin", CGB2ArgMin]
]

inferenceRule[ CGB2ArgMax] =
this:PRFSIT$[ g_FML$, KB_List, _, rest___?OptionQ] :> performProofStep[
	abbreviateArgminmax[ this, ArgMax$TM, "ArgMax", CGB2ArgMax]
]
	
abbreviateArgminmax[ ps:PRFSIT$[ g_FML$, KB_List, _, rest___], fun_, optsName_String, ruleName_] :=
	Module[ {usedExpressions, pos, expr = $Failed, cond1, cond2, assm1, assm2, assm3, newK, gen, newG, tmp, dom, rng, cnd, term, abf, locC, var},
		usedExpressions = getOptionalComponent[ ps, optsName];
		pos = Position[ g, E:(fun[ RNG$[ _], _, d_, _])
								/; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTOM$TM|Theorema`Knowledge`isTotallyOrdered$TM)[ d], __]] && !MemberQ[ usedExpressions, E] && freeVariables[ E] === {},
						Infinity, 1
			  ];
		If[ pos === {},
			pos = Position[ KB, E:(fun[ RNG$[ _], _, d_, _])
									/; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTOM$TM|Theorema`Knowledge`isTotallyOrdered$TM)[ d], __]] && !MemberQ[ usedExpressions, E] && freeVariables[ E] === {},
							Infinity, 1
				  ];
			If[ pos =!= {}, expr = myExtract[ KB, First[ pos]]],
			expr = myExtract[ g, First[ pos]];
		];
		If[ expr === $Failed,
			$Failed,
			
			dom = expr[[ 3]];
			rng = expr[[ 1]];
			cnd = expr[[ 2]];
			term = expr[[ 4]];
			var = First[ rngVariables[ rng]];
			locC = getOptionalComponent[ ps, "constants"];
			abf = newABF[ First[ var], Prepend[ KB, g]];
			cond1 = makeConditionFML[ formula -> Exists$TM[ rng, True, cnd]];
			cond2 = makeConditionFML[ formula -> Forall$TM[ rng, cnd, DomainOperation$TM[ dom, Element$TM][ term]]];
			assm1 = makeAssumptionFML[ formula -> substituteFree[ First[ myRngToCondition[ rng]], {var -> abf}]];
			assm2 = makeAssumptionFML[ formula -> substituteFree[ cnd, {var -> abf}]];
			assm3 = makeAssumptionFML[ formula -> Forall$TM[ rng, cnd,
															Switch[ fun,
																ArgMin$TM,
																DomainOperation$TM[ dom, LessEqual$TM][ substituteFree[ term, {var -> abf}], term],
																_,
																DomainOperation$TM[ dom, LessEqual$TM][ term, substituteFree[ term, {var -> abf}]]
															]
													]
					];
			newK = joinNewKB[ {}, {cond2, assm1, assm2, assm3}];
			gen = {cond1, {cond2, assm1, assm2, assm3}};
			newG = formula@g /. expr -> abf;
			If[ newG === formula@g,
				newG = g,
				newG = makeGoalFML[ formula -> newG]
			];
			AppendTo[ gen, {newG}];
			Scan[ Function[ fml,
					tmp = formula@fml /. expr -> abf;
					If[ tmp === formula@fml,
						newK = addNewKB[ newK, fml],
						tmp = makeAssumptionFML[ formula -> tmp];
						newK = addNewKB[ newK, tmp];
						AppendTo[ gen, tmp]
					]
				],
				KB
			];
			PrependTo[ usedExpressions, expr];
			makeANDNODE[ makePRFINFO[ name -> ruleName, used -> {{expr, abf}}, generated -> {gen}],
						{Block[ {$rewriteRules = {}}, toBeProved[ goal -> cond1, kb -> KB, optsName -> usedExpressions, rest]],
						Block[ {$rewriteRules = {}}, toBeProved[ goal -> cond2, kb -> KB, optsName -> usedExpressions, rest]],
						toBeProved[ goal -> newG, kb -> newK, "constants" -> Prepend[ locC, substituteFree[ rng, {var -> abf}]], rest]}]
		]
	]


(* ::Section:: *)
(* min, max *)

inferenceRule[ CGB2tupleAdditionGoal] =
PRFSIT$[ g:FML$[ _, form_, __], KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {pos, newG, expr = $Failed},
		pos = Position[ form, DomainOperation$TM[ dom_, Plus$TM][ _, DomainOperation$TM[ dom_, min$TM|max$TM][ _Tuple$TM]]
								/; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]], Infinity, 1];
		If[ pos === {},
			pos = Position[ form, DomainOperation$TM[ dom_, Plus$TM][ DomainOperation$TM[ dom_, min$TM|max$TM][ _Tuple$TM], _]
									/; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]], Infinity, 1];
			If[ pos === {},
				pos = Position[ form, DomainOperation$TM[ dom_, Plus$TM][ _, DomainOperation$TM[ dom_, min$TM|max$TM][ _TupleOf$TM]]
										/; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]], Infinity, 1];
				If[ pos === {},
					pos = Position[ form, DomainOperation$TM[ dom_, Plus$TM][ DomainOperation$TM[ dom_, min$TM|max$TM][ _TupleOf$TM], _]
											/; MemberQ[ KB, FML$[ _, Theorema`Knowledge`isTOM$TM[ dom], __]], Infinity, 1];
					If[ pos =!= {},
						pos = First[ pos];
						expr = Replace[ myExtract[ form, pos],
										(h:(DomainOperation$TM[ dom_, Plus$TM]))[ (f:(DomainOperation$TM[ dom_, min$TM|max$TM]))[ TupleOf$TM[ r_RNG$, cond_, term_]], x_] :>
										f[ TupleOf$TM[ r, cond, h[ term, x]]]
								]
					],
					pos = First[ pos];
					expr = Replace[ myExtract[ form, pos],
									(h:(DomainOperation$TM[ dom_, Plus$TM]))[ x_, (f:(DomainOperation$TM[ dom_, min$TM|max$TM]))[ TupleOf$TM[ r_RNG$, cond_, term_]]] :>
									f[ TupleOf$TM[ r, cond, h[ x, term]]]
							]
				],
				pos = First[ pos];
				expr = Replace[ myExtract[ form, pos],
								(h:(DomainOperation$TM[ dom_, Plus$TM]))[ (f:(DomainOperation$TM[ dom_, min$TM|max$TM]))[ t_], x_] :>
								f[Map[ h[ #, x]&, t]]
						]
			],
			pos = First[ pos];
			expr = Replace[ myExtract[ form, pos],
							(h:(DomainOperation$TM[ dom_, Plus$TM]))[ x_, (f:(DomainOperation$TM[ dom_, min$TM|max$TM]))[ t_]] :>
							f[Map[ h[ x, #]&, t]]
					]
		];
		If[ expr === $Failed,
			$Failed,
			newG = makeGoalFML[ formula -> myReplacePart[ form, pos -> expr]];
			makeANDNODE[ makePRFINFO[ name -> CGB2tupleAdditionGoal, used -> g, generated -> newG],
						toBeProved[ goal -> newG, kb -> KB, rest]
			]
		]
	]
]

inferenceRule[ CGB2minGoalLeft] =
PRFSIT$[ g:FML$[ _, (h:(DomainOperation$TM[ dom_, Less$TM|LessEqual$TM]))[ DomainOperation$TM[ dom_, min$TM][ S_], x_], __], KB_List, _, rest___?OptionQ] :>
performProofStep[
	Module[ {newG},
		newG = makeGoalFML[ formula -> And$TM[ DomainOperation$TM[ dom, Theorema`Language`SquareSubsetEqual$TM][ S],
											   Exists$TM[ RNG$[ STEPRNG$[ VAR$[ Theorema`Knowledge`i$TM], 1, BracketingBar$TM[ S], 1]],
											   				True,
											   				h[ Subscript$TM[ S, VAR$[ Theorema`Knowledge`i$TM]], x]
											   	]
										]
				];
		makeANDNODE[ makePRFINFO[ name -> CGB2minGoalLeft, used -> g, generated -> newG],
					toBeProved[ goal -> newG, kb -> KB, rest]
		]
	]
] /; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTOM$TM|Theorema`Knowledge`isTotallyOrdered$TM)[ dom], __]]

inferenceRule[ CGB2maxGoalRight] =
PRFSIT$[ g:FML$[ _, (h:(DomainOperation$TM[ dom_, Less$TM|LessEqual$TM]))[ x_, DomainOperation$TM[ dom_, max$TM][ S_]], __], KB_List, _, rest___?OptionQ] :>
performProofStep[
	Module[ {newG},
		newG = makeGoalFML[ formula -> And$TM[ DomainOperation$TM[ dom, Theorema`Language`SquareSubsetEqual$TM][ S],
											   Exists$TM[ RNG$[ STEPRNG$[ VAR$[ Theorema`Knowledge`i$TM], 1, BracketingBar$TM[ S], 1]],
											   				True,
											   				h[ x, Subscript$TM[ S, VAR$[ Theorema`Knowledge`i$TM]]]
											   	]
										]
				];
		makeANDNODE[ makePRFINFO[ name -> CGB2maxGoalRight, used -> g, generated -> newG],
					toBeProved[ goal -> newG, kb -> KB, rest]
		]
	]
] /; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTOM$TM|Theorema`Knowledge`isTotallyOrdered$TM)[ dom], __]]

inferenceRule[ CGB2minKBRight] =
PRFSIT$[ g_FML$, KB:{pre___FML$, form:FML$[ _, (h:(DomainOperation$TM[ dom_, Less$TM|LessEqual$TM]))[ x_, DomainOperation$TM[ dom_, min$TM][ S_]], __], post___FML$}, _, rest___?OptionQ] :>
performProofStep[
	Module[ {newForm, domCond},
		domCond = makeConditionFML[ formula -> DomainOperation$TM[ dom, Theorema`Language`SquareSubsetEqual$TM][ S]];
		newForm = makeAssumptionFML[ formula -> Forall$TM[ RNG$[ STEPRNG$[ VAR$[ Theorema`Knowledge`i$TM], 1, BracketingBar$TM[ S], 1]],
											   				True,
											   				h[ x, Subscript$TM[ S, VAR$[ Theorema`Knowledge`i$TM]]]
										]
				];
		makeANDNODE[ makePRFINFO[ name -> CGB2minKBRight, used -> {form, g}, generated -> {domCond, newForm}],
					{Block[ {$rewriteRules = {}}, toBeProved[ goal -> domCond, kb -> {pre, post}, rest]],
					toBeProved[ goal -> g, kb -> joinNewKB[ {pre, post}, {domCond, newForm}, Equality -> Weak, Mode -> Append], rest]}
		]
	]
] /; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTOM$TM|Theorema`Knowledge`isTotallyOrdered$TM)[ dom], __]]

inferenceRule[ CGB2maxKBLeft] =
PRFSIT$[ g_FML$, KB:{pre___FML$, form:FML$[ _, (h:(DomainOperation$TM[ dom_, Less$TM|LessEqual$TM]))[ DomainOperation$TM[ dom_, max$TM][ S_], x_], __], post___FML$}, _, rest___?OptionQ] :>
performProofStep[
	Module[ {newForm, domCond},
		domCond = makeConditionFML[ formula -> DomainOperation$TM[ dom, Theorema`Language`SquareSubsetEqual$TM][ S]];
		newForm = makeAssumptionFML[ formula -> Forall$TM[ RNG$[ STEPRNG$[ VAR$[ Theorema`Knowledge`i$TM], 1, BracketingBar$TM[ S], 1]],
											   				True,
											   				h[ Subscript$TM[ S, VAR$[ Theorema`Knowledge`i$TM]], x]
										]
				];
		makeANDNODE[ makePRFINFO[ name -> CGB2maxKBLeft, used -> {form, g}, generated -> {domCond, newForm}],
					{Block[ {$rewriteRules = {}}, toBeProved[ goal -> domCond, kb -> {pre, post}, rest]],
					toBeProved[ goal -> g, kb -> joinNewKB[ {pre, post}, {domCond, newForm}, Equality -> Weak, Mode -> Append], rest]}
		]
	]
] /; MemberQ[ KB, FML$[ _, (Theorema`Knowledge`isTOM$TM|Theorema`Knowledge`isTotallyOrdered$TM)[ dom], __]]


(* ::Section:: *)
(* Quantifiers *)

inferenceRule[ CGB2forallGoal] = 
ps:PRFSIT$[ g:FML$[ _, u:Forall$TM[ rng_, cond_, A_], __], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {simp, rc, r, c, f, fix, newConds, newGoal, locC},
		simp = computeInProof[ u];
		If[ MatchQ[ simp, _Forall$TM],
			rc = myRngToCondition[ rng];
			If[ !FreeQ[ rc, $Failed], 
				$Failed,
				(* else *)
				{{r, c, f}, fix} = arbitraryButFixed[ {rc, cond, A}, rng, {g, k}];
				locC = getOptionalComponent[ ps, "constants"];
				newGoal = makeGoalFML[ formula -> f];
				newConds = Map[ makeAssumptionFML[ formula -> #]&, DeleteCases[ Append[ r, c], True]];
				makeANDNODE[ makePRFINFO[ name -> forallGoal, used -> g, "abf" -> rngConstants[ fix]], 
					toBeProved[ goal -> newGoal, kb -> joinNewKB[ k, newConds, Mode -> Prepend, Equality -> Weak], "constants" -> Prepend[ locC, fix], rest]]
			],
			(* else *)
			simp = makeGoalFML[ formula -> simp];
			makeANDNODE[ makePRFINFO[ name -> forallGoal, used -> g], 
				toBeProved[ goal -> simp, kb -> k, rest]]
		]
	]
]

inferenceRule[ CGB2existsGoal] = 
ps:PRFSIT$[ g:FML$[ _, u:Exists$TM[ rng_, cond_, A_], __], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {simp, rc, rv, conjuncts, patterns, newG, out = $Failed, subst, subst2, currentSubst, instRng, thinnedRng},
		simp = computeInProof[ u];
		If[ MatchQ[ simp, _Exists$TM],
			rv = rngVariables[ rng];
			rc = DeleteCases[ DeleteDuplicates[ myRngToCondition[ rng]], True];
			If[ FreeQ[ rc, $Failed],
				If[ MemberQ[ rc, False],
					newG = makeGoalFML[ formula -> False];
					out = makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> g, generated -> newG, "instantiation" -> {}], 
						toBeProved[ goal -> newG, kb -> k, rest]],
							
					conjuncts = DeleteCases[ DeleteDuplicates[ Apply[ List, Flatten[ And$TM[ cond, A], Infinity, And$TM]]], True];
					Which[ MemberQ[ conjuncts, False],
						newG = makeGoalFML[ formula -> False];
						out = makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> g, generated -> newG, "instantiation" -> {}], 
							toBeProved[ goal -> newG, kb -> k, rest]],
							
						conjuncts === {},
						If[ rc === {},
							newG = makeGoalFML[ formula -> True];
							out = makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> g, generated -> newG, "instantiation" -> {}], 
								toBeProved[ goal -> newG, kb -> k, rest]],
							
							patterns = Map[ existentialPattern[ #, rv]&, rc]
						],
						
						True,
						patterns = Map[ existentialPattern[ #, rv]&, conjuncts];
					]
				];
				If[ out === $Failed,
					subst = Map[ Function[ p,
									currentSubst = {};
									Scan[ (currentSubst = Join[ currentSubst, ReplaceList[ formula@#, p]])&, k];
									currentSubst
								],
								patterns
							];
					(* 'subst' is now a (non-empty) list of (possibly empty) lists of substitutions,
					where the i-th list contains substitutions that witness the i-th conjunct *)
					subst = Fold[ simpleUnify, First[ subst], Rest[ subst]];
					(* 'subst' is now a (possibly empty) list of substitutions that witness the whole conjunction *)
					If[ subst =!= {},
						If[ conjuncts === {},
							(* In this case, we have an existential quantifier with trivial 'A' and 'cond', and
							we managed to find formulae witnessing all range conditions -> done. *)
							newG = makeGoalFML[ formula -> True];
							out = makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> g, generated -> newG, "instantiation" -> First[ subst]], 
									toBeProved[ goal -> newG, kb -> k, rest]],
							
							(* else *)
							(* In this case, we managed to find witnesses for 'A' and 'cond'; Hence, we try to find
							some for the range conditions as well *)
							patterns = Map[ existentialPattern[ #, rv]&, rc];
							subst2 = Map[ Function[ p,
										currentSubst = {};
										Scan[ (currentSubst = Join[ currentSubst, ReplaceList[ formula@#, p]])&, k];
										currentSubst
									],
									patterns
								];
							subst2 = Fold[ simpleUnify, subst, subst2];
							
							If[ subst2 === {},
								(* We just take the first substitution that leads to a witness for 'A' and 'cond'.
								This is maybe not the best strategy ... *)
								subst = First[ subst];
								instRng = Select[ rng, MemberQ[ Map[ First, subst], First[#]]&];
								thinnedRng = Complement[ rng, instRng];
					            rc = myRngToCondition[ instRng];
					            If[ FreeQ[ rc, $Failed],				
					            	newG = makeGoalFML[ formula -> makeExist[ thinnedRng, rc, True, True, subst]];
					            	out = makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> {{g}}, generated -> {{newG}}, "instantiation" -> subst], 
					                	toBeProved[ goal -> newG, kb -> k, rest]]
					            ],
					            
								(* else *)
								(* In this case we managed to find witnesses for everything -> done. *)
								newG = makeGoalFML[ formula -> True];
								out = makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> {{g}}, generated -> {{newG}}, "instantiation" -> First[ subst2]], 
										toBeProved[ goal -> newG, kb -> k, rest]]
							]
						]
					]
				];
				out,
				(* else *)
				$Failed
			],
			simp = makeGoalFML[ formula -> simp];
			makeANDNODE[ makePRFINFO[ name -> CGB2existsGoal, used -> g], 
				toBeProved[ goal -> simp, kb -> k, rest]]
		]
	]
]

inferenceRule[ CGB2existsKB] = 
ps:PRFSIT$[ g_, k:{pre___, e:FML$[ _, u:Exists$TM[ rng_, cond_, A_], __], post___}, id_, rest___?OptionQ] :> performProofStep[
	Module[ {simp, r, c, f, fix, newConds, locC},
		simp = computeInProof[ u];
		If[ MatchQ[ simp, _Exists$TM],
			(* no simplification *)
			{{r, c, f}, fix} = arbitraryButFixed[ {myRngToCondition[ rng], cond, A}, rng, {g, k}];
			locC = Prepend[ getOptionalComponent[ ps, "constants"], fix];
			newConds = Map[ makeAssumptionFML[ formula -> #]&, DeleteCases[ Join[ r, {c, f}], True]];
			makeANDNODE[ makePRFINFO[ name -> existsKB, used -> e, "abf" -> rngConstants[ fix]], 
				toBeProved[ goal -> g, kb -> joinNewKB[ {pre, post}, newConds, Mode -> Prepend, Equality -> Weak], "constants" -> locC, rest]],
			(* else *)
			simp = makeAssumptionFML[ formula -> simp];
			makeANDNODE[ makePRFINFO[ name -> existsKB, used -> e], 
				toBeProved[ goal -> g, kb -> addNewKB[ {pre, post}, simp, Mode -> Prepend, Equality -> Weak], rest]]
		]
	]
]

inferenceRule[ CGB2forallKBInteractive] = 
ps:PRFSIT$[ g_, k_List?(MemberQ[ #, FML$[ _, Forall$TM[ _, _, _], ___]]&), id_, rest___?OptionQ] :> performProofStep[
    Module[ {fml, rc, instRng, thinnedRng, fInst, sel, rng, cond, A, subst},
    	sel = selectAssumptionDialog[ g, k, FML$[ _, Forall$TM[ _, _, _], __], "selectForallHeader", "!selectForall"];
        NotebookClose[ $TMAproofNotebook];
        If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ k],
        	fml = k[[sel]];
        	If[ MatchQ[ fml, FML$[ _, Forall$TM[ _, _, _], __]],
	        	rng = First[ formula@fml];
		        If[ FreeQ[ rc, $Failed],
		            subst = instantiateForallKBInteractive[ getOptionalComponent[ ps, "constants"] /. RNG$ -> Sequence, rng, {g, k}, key@fml];
		            If[ subst === $Failed,
		                $Failed,
		                cond = (formula@fml)[[2]];
			        	A = Last[ formula@fml];
			        	instRng = Select[ rng, MemberQ[ Map[ First, subst], First[#]]&];
						thinnedRng = Complement[ rng, instRng];
			            rc = myRngToCondition[ instRng];
			            If[ FreeQ[ rc, $Failed],
			                fInst = makeAssumptionFML[ formula -> makeForall[ thinnedRng, rc, cond, A, subst]];
			                makeANDNODE[ makePRFINFO[ name -> forallKBInteractive, used -> fml, "instantiation" -> subst], 
			                    toBeProved[ goal -> g, kb -> addNewKB[ k, fInst, Mode -> Prepend, Equality -> Weak], rest]],
			                $Failed
			            ]
		            ],
		            $Failed
		        ],
		        $Failed
        	],
        	$Failed
        ]
    ]
]
makeForall[ RNG$[], cond1_List, cond2_, A_, subst_] :=
	Module[ {cond = substituteFree[ DeleteCases[ Append[ cond1, cond2], True], subst],
			fml = substituteFree[ A, subst]},
		If[ cond === {},
			fml,
			Implies$TM[ makeConjunction[ cond, And$TM], fml]
		]
	]
makeForall[ r:RNG$[__], cond1_List, cond2_, A_, subst_] := 
	Forall$TM[ substituteFree[ r, subst], True, makeForall[ RNG$[], cond1, cond2, A, subst]]
	
(* We want to have our own dialog, that's the only reason for implementing the inference rule ourselves. *)
inferenceRule[ CGB2existsGoalInteractive] = 
ps:PRFSIT$[ g:FML$[ _, u:Exists$TM[ rng_, cond_, A_], __], k_List, id_, rest___?OptionQ] :> performProofStep[
    Module[ {subst, rc, instRng, thinnedRng, newGoal},
    	(* Note the 's' in 'instantiateExist_s_GoalInteractive', making the function different from the public one!*)
        subst = instantiateExistsGoalInteractive[ g, getOptionalComponent[ ps, "constants"] /. RNG$ -> Sequence, k];
        If[ subst === $Failed,
        	(* the interactive dialog has been canceled or closed *)
            $Failed,
            (* else *)
            instRng = Select[ rng, MemberQ[ Map[ First, subst], First[#]]&];
			thinnedRng = Complement[ rng, instRng];
            rc = myRngToCondition[ instRng];
            If[ !FreeQ[ rc, $Failed], 
				$Failed,
				(* else *)				
            	newGoal = makeGoalFML[ formula -> makeExist[ thinnedRng, rc, cond, A, subst]];
            	makeANDNODE[ makePRFINFO[ name -> existsGoalInteractive, used -> g, "instantiation" -> subst], 
                	toBeProved[ goal -> newGoal, kb -> k, rest]]
            ]
        ]
    ]
]
makeExist[ RNG$[], cond1_List, cond2_, A_, subst_] :=
	Module[ { c = substituteFree[ DeleteCases[ Join[ cond1, {cond2, A}], True], subst]},
		If[ c === {},
			True,
			Apply[ And$TM, c]
		]
	]
makeExist[ r:RNG$[__], cond1_List, cond2_, A_, subst_] := 
	Exists$TM[ substituteFree[ r, subst], True, makeExist[ RNG$[], cond1, cond2, A, subst]]
	
	
inferenceRule[ CGB2existsTupleOfGoal] =
PRFSIT$[ g:FML$[ _, Exists$TM[ RNG$[ STEPRNG$[ j_VAR$, 1, BracketingBar$TM[ T:(TupleOf$TM[ r_RNG$, phi_, t_])], 1]], psi_, chi_], __],
		KB_List, _, rest___?OptionQ] :> performProofStep[
	Module[ {newG},
		newG = makeGoalFML[ formula -> And$TM[ isTuple$TM[ T], Exists$TM[ r, phi, And$TM[ psi, chi] /. Subscript$TM[ T, j] :> t]]];
		makeANDNODE[ makePRFINFO[ name -> CGB2existsTupleOfGoal, used -> g, generated -> newG], 
                	toBeProved[ goal -> newG, kb -> KB, rest]]
	]
] /; !(MemberQ[ freeVariables[ psi /. Subscript$TM[ T, j] :> Null], j] || MemberQ[ freeVariables[ chi /. Subscript$TM[ T, j] :> Null], j])

inferenceRule[ CGB2forallTupleOfKB] =
PRFSIT$[ g_FML$, KB:{KBa___, k:FML$[ _, Forall$TM[ RNG$[ STEPRNG$[ j_VAR$, 1, BracketingBar$TM[ T:(TupleOf$TM[ r_RNG$, phi_, t_])], 1]], psi_, chi_], __], KBb___}, _, rest___?OptionQ] :>
performProofStep[
	Module[ {newg1, newk},
		newg1 = makeConditionFML[ formula -> isTuple$TM[ T]];
		newk = makeAssumptionFML[ formula -> Forall$TM[ r, And$TM[ phi, psi] /. Subscript$TM[ T, j] :> t, chi /. Subscript$TM[ T, j] :> t]];
		makeANDNODE[ makePRFINFO[ name -> CGB2forallTupleOfKB, used -> {k, g}, generated -> {newg1, newk}],
					{Block[ {$rewriteRules = {}}, toBeProved[ goal -> newg1, kb -> {KBa, KBb}, rest]],
					toBeProved[ goal -> g, kb -> joinNewKB[ {KBa, KBb}, {newg1, newk}, Equality -> Weak, Mode->Append], rest]}
		]
	]
] /; !(MemberQ[ freeVariables[ psi /. Subscript$TM[ T, j] :> Null], j] || MemberQ[ freeVariables[ chi /. Subscript$TM[ T, j] :> Null], j])

inferenceRule[ CGB2existsAppendGoal] =
PRFSIT$[ g:FML$[ _, Exists$TM[ RNG$[ STEPRNG$[ j_VAR$, 1, (BracketingBar$TM[ appendElem$TM[ A_, _]])|
						(DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Plus$TM][ BracketingBar$TM[ A_], 1]),
					1]], psi_, chi_], __],
		KB_List, _, rest___?OptionQ] :>
performProofStep[
	Module[ {gen, tuple, element},
		element = Cases[ g, T:(appendElem$TM[ A, e_]) :> {T, e} /;
								!(MemberQ[ freeVariables[ psi /. Subscript$TM[ T, j] :> Null], j] ||
									MemberQ[ freeVariables[ chi /. Subscript$TM[ T, j] :> Null], j]),
						Infinity, 1
				];
		If[ Length[ element] === 1,
			{{tuple, element}} = element;
			gen = makeGoalFML[ formula -> And$TM[
					isTuple$TM[ A],
					Or$TM[
						Exists$TM[ RNG$[ STEPRNG$[ j, 1, BracketingBar$TM[ A], 1]], psi /. {Subscript$TM[ tuple, j] :> Subscript$TM[ A, j]}, chi /. {Subscript$TM[ tuple, j] :> Subscript$TM[ A, j]}],
						And$TM[ psi, chi] /. {Subscript$TM[ tuple, j] :> element}
					]
				  ]
				];
			makeANDNODE[ makePRFINFO[ name -> CGB2existsAppendGoal, used -> g, generated -> gen],
						toBeProved[ goal -> gen, kb -> KB, rest]
			],
			
			$Failed
		]
	]
]


inferenceRule[ CGB2forallAppendKB] =
PRFSIT$[ g_FML$, KB:{___, form:FML$[ _, Forall$TM[ RNG$[ STEPRNG$[ j_VAR$, 1, (BracketingBar$TM[ appendElem$TM[ A_, _]])|
						(DomainOperation$TM[ IntegerInterval$TM[ DirectedInfinity[-1], DirectedInfinity[1], False, False], Plus$TM][ BracketingBar$TM[ A_], 1]),
					1]], psi_, chi_], __], ___},
		_, rest___?OptionQ] :>
Module[ {gen, newg, tuple, element},
	Block[ {$rewriteRules = {}, $generated = {}},
		{{tuple, element}} = element;
		newg = makeConditionFML[ formula -> isTuple$TM[ A]];
		gen = makeAssumptionFML[ formula -> And$TM[
				Forall$TM[ RNG$[ STEPRNG$[ j, 1, BracketingBar$TM[ A], 1]], psi /. {Subscript$TM[ tuple, j] :> Subscript$TM[ A, j]}, chi /. {Subscript$TM[ tuple, j] :> Subscript$TM[ A, j]}],
				Implies$TM[ psi, chi] /. {Subscript$TM[ tuple, j] :> element}
			  ]
			];
		makeANDNODE[ makePRFINFO[ name -> CGB2forallAppendKB, used -> {form, g}, generated -> {newg, gen}],
					{toBeProved[ goal -> newg, kb -> DeleteCases[ KB, form], rest],
					toBeProved[ goal -> g, kb -> joinNewKB[ DeleteCases[ KB, form], {gen, newg}, Equality -> Weak, Mode -> Append], rest]}
		]
	] /; (
		element = Cases[ form, T:(appendElem$TM[ A, e_]) :> {T, e} /;
							!(MemberQ[ freeVariables[ psi /. Subscript$TM[ T, j] :> Null], j] ||
								MemberQ[ freeVariables[ chi /. Subscript$TM[ T, j] :> Null], j]),
						Infinity, 1
				];
		Length[ element] === 1
	)
]



(* ::Section:: *)
(* Implications *)

(* TODO: Possible improvement: Remove rewrite rules corresponding to "old" formula, because they still have
	conditions that are not necessary any longer. *)
inferenceRule[ CGB2modusPonensTollens] =
this:PRFSIT$[ g_, k:{__FML$}?((!FreeQ[ #, Implies$TM[ _, _]])&), id_, rest___?OptionQ] :> performProofStep[
	Module[ {f, pos, impl, assm, gRules, condRefs = {}, newForm, tmpRefs, action},
		gRules = filterRules[ goalRules@this, None];
		pos = Position[ k, Implies$TM[ _, _], Heads -> False];
		Catch[
			Scan[ (
				f = Part[ k, First[ #]];
				impl = Extract[ k, #];
				assm = First[ impl];
				Switch[ assm,
					
					And$TM[ __],
					newForm = And$TM[];
					condRefs = {};
					Scan[ Function[ conj,
						tmpRefs = followsImmediately[ conj, gRules];
						If[ tmpRefs === False,
							AppendTo[ newForm, conj],
							condRefs = Join[ condRefs, tmpRefs]
						]
						],
						assm
					];
					newForm = Which[ newForm === And$TM[],
									action = "modusPonens";
									makeAssumptionFML[ formula -> myReplacePart[ formula@f, Drop[ #, 2] -> Last[ impl]]],
									
									condRefs =!= {},
									action = "simplification";
									makeAssumptionFML[ formula -> ReplacePart[ formula@f, Append[ Drop[ #, 2], 1] -> newForm]],
									
									True,
									$Failed
								],
					
					_,
					condRefs = followsImmediately[ assm, gRules];
					action = "modusPonens";
					newForm = If[ condRefs === False,
								$Failed,
								makeAssumptionFML[ formula -> myReplacePart[ formula@f, Drop[ #, 2] -> Last[ impl]]]
								]
				];
				If[ newForm === $Failed,
					assm = Not$TM[ Last[ impl]];
					condRefs = followsImmediately[ assm, gRules];
					If[ condRefs =!= False,
						newForm = makeAssumptionFML[ formula -> myReplacePart[ formula@f, Drop[ #, 2] -> sinkNegation[ Not$TM[ First[ impl]]]]];
						action = "modusTollens"
					]
				];
				If[ newForm =!= $Failed,
					Throw[ makeANDNODE[ makePRFINFO[
												name -> CGB2modusPonensTollens,
												used -> f,
												generated -> newForm,
												"usedForms" -> DeleteDuplicates[ condRefs],
												"action" -> action], 
								toBeProved[ goal -> g, kb -> addNewKB[ Delete[ k, First[ #]], newForm, Mode -> Prepend, Equality -> Weak], rest]]
					]
				]
				)&,
				pos
			];
			$Failed
		]
	]
]
	
inferenceRule[ CGB2selectImplicationInteractive] = 
PRFSIT$[ g_FML$, k_List?(MemberQ[ #, FML$[ _, Implies$TM[ _, _], ___]]&), id_, rest___?OptionQ] :> performProofStep[
	Module[ {sel, fml, assm, concl, assmGoal, negGoal},
		sel = selectAssumptionDialog[ g, k, FML$[ _, Implies$TM[ _, _], ___], "selectImplicationHeader", "!selectImplication"];
		NotebookClose[ $TMAproofNotebook];
		If[ IntegerQ[ sel] && sel >= 1 && sel <= Length[ k],
			fml = k[[sel]];
			assm = makeAssumptionFML[ formula -> First[ formula@fml]];
			concl = makeAssumptionFML[ formula -> Last[ formula@fml]];
			assmGoal = makeGoalFML[ formula -> First[ formula@fml]];
			negGoal = makeAssumptionFML[ formula -> sinkNegation[ Not$TM[ formula@g]]];
			makeANDNODE[ makePRFINFO[ name -> CGB2selectImplicationInteractive, used -> {fml, g}, generated -> {{assmGoal, negGoal}, {assm, concl}}], 
                	{toBeProved[ goal -> assmGoal,
                				kb -> addNewKB[ k, negGoal, Mode -> Prepend, Equality -> Weak],
                				rest],
                	 Block[ {$rewriteRules = {}}, toBeProved[ goal -> g,
                				kb -> joinNewKB[ Delete[ k, sel], {assm, concl}, Mode -> Prepend, Equality -> Weak],
                				rest]]
                	}
            ],
			$Failed
		]
	]
]


(* ::Section:: *)
(* Rewriting *)

inferenceRule[ CGB2generateRewriteRules] = 
this:PRFSIT$[ g_, k_List, id_, opts___?OptionQ] :> performProofStep[
	Module[ { newKB = {}, newOpts, form, act},
		(* Formulae in the initial proof situation are, for some reason, not simplified.
		That's why we have to take care of it ourselves. *)
		Scan[ (
			form = simplifyFormulaByComputation[ #];
			newKB = addNewKB[ newKB, form, Equality -> Weak]
			)&,
			k
		];
		
		(* We also have to take into account all formulae that are not in the KB any more because they have already
		been transformed into rewrite rules. This is somewhat tedious, but probably there is no better way ... *)
		Scan[ (
			form = simplifyFormulaByComputation[ extractFormulaFromRule[ #]];
			newKB = addNewKB[ newKB, form, Equality -> Weak]
			)&,
			Join[ kbRules@this, goalRules@this, substRules@this, defRules@this]
		];
		
		act = ruleActivity@this /. {(h:(Rule|RuleDelayed))[ CGB2generateRewriteRules, _] :> h[ CGB2generateRewriteRules, False]};
		
		newOpts = {opts} /. {(h:(Rule|RuleDelayed))[ r:(kbRules|goalRules|substRules|defRules), _] :> h[ r, {}],
							(h:(Rule|RuleDelayed))[ ruleActivity, _] :> h[ ruleActivity, act]};
		makeANDNODE[ makePRFINFO[ name -> CGB2generateRewriteRules, used -> {}, generated -> {}], 
			toBeProved[ goal -> simplifyFormulaByComputation[ g], kb -> newKB, Apply[ Sequence, newOpts]]]
	]
]

(* The following inference rule is almost completely the same as the one in "BasicTheoremaLanguageRules",
	with the following differences:
	- the condition on the goal being atomic is dropped
	- no book-keeping concerning rules that have already been tried is done
	- if no rewrite is possible, $Failed is returned
	- the goal is only rewritten to the "simplest" formulae (according to function "simplestFormulae")
*)
inferenceRule[ CGB2goalRewriting] = 
this:PRFSIT$[ g:FML$[ _, f_, __], k_List, id_, rest___?OptionQ] :> performProofStep[
	Module[ {usedSubsts, conds, newForms, simplifiedForms = {}, j, newNodes = {}},
		{newForms, usedSubsts, conds} = replaceListAndTrack[ formula@g, filterRules[ goalRules@this, key@g]];
		Do[
			AppendTo[ simplifiedForms, {makeGoalFML[ formula -> newForms[[j]]], usedSubsts[[j]]}];
			(* Goal rewriting should actually generate NO conditions. If a condition still appears, there must have gone something wrong *)
			Assert[ conds[[j]]],
			{j, Length[ newForms]}
		];
		newNodes = Map[ makeANDNODE[
							makePRFINFO[ name -> goalRewriting, used -> Prepend[ Last[ #], g], generated -> First[ #]],
							toBeProved[ goal -> First[ #], kb -> k, rest]
						]&,
						simplestFormulae[ simplifiedForms]
					];
		Switch[ Length[ newNodes],
			0,
			$Failed,
			1,
			First[ newNodes],
			_,
			makeORNODE[ 
				makePRFINFO[ name -> multipleGoalRewriting, used -> used@newNodes, generated -> generated@newNodes],
				newNodes]
		]            	
	]
]

(* The following are the differences between "CGB2elementarySubstitution" and "elementarySubstitution":
-	Rather than "replaceRepeatedAndTrack" the new function "replaceFirstPossible" is used
-	Hence, no conditions have to be checked afterwards, i.e. only one new proof situation will be generated
-	Also, "CGB2elementarySubstitution" only applies rules with variables in their LHS. "CGB2groundSubstitution"
	applies all other rules (with ground LHS).
In fact, "CGB2elementarySubstitution" and "CGB2expandDef" do almost exactly the same: The only difference is that
"CGB2elementarySubstitution" uses (non-ground) elementary substitution rules and "CGB2expandDef" uses definition rules.
*)
inferenceRule[ CGB2elementarySubstitution] = 
ps:PRFSIT$[ g_, k_List, id_, rest___?OptionQ] :> performProofStep[
	rewriteProofSit[
		g,
		k,
		Cases[ substRules@ps, {_, (Rule|RuleDelayed)[ _?(!FreeQ[ #, _Blank|_BlankSequence|_BlankNullSequence]&), _]}],
		filterRules[ goalRules@ps, None],
		CGB2elementarySubstitution,
		rest
	]
]

inferenceRule[ CGB2groundSubstitution] = 
ps:PRFSIT$[ g_, k_List, id_, rest___?OptionQ] :> performProofStep[
	rewriteProofSit[
		g,
		k,
		Cases[ substRules@ps, {_, (Rule|RuleDelayed)[ _?(FreeQ[ #, _Blank|_BlankSequence|_BlankNullSequence]&), _]}],
		filterRules[ goalRules@ps, None],
		CGB2groundSubstitution,
		rest
	]
]

(* The following are the differences between "CGB2expandDef" and "expandDef":
-	Rather than "replaceAllAndTrack" the new function "replaceFirstPossible" is used
-	Hence, no conditions have to be checked afterwards, i.e. only one new proof situation will be generated
*)
inferenceRule[ CGB2expandDef] = 
ps:PRFSIT$[ g_, k_List, id_, rest___?OptionQ] :> performProofStep[
	rewriteProofSit[ g, k, defRules@ps, filterRules[ goalRules@ps, None], CGB2expandDef, rest]
]

rewriteProofSit[ _, _List, {}, _List, _, ___?OptionQ] := $Failed
rewriteProofSit[ g_, k_List, rules_List, gRules_List, n_, rest___?OptionQ] :=
	Module[ {usedDef, condDefs, new, newG, newForm, newK = {}, rewritten = False, j, usedForms, genForms, replBy = {}, eh},
		{new, usedDef, condDefs} = replaceFirstPossible[ formula@g, filterRules[ rules, None], gRules];
		If[ usedDef =!= {},
			newG = makeGoalFML[ formula -> new];
			rewritten = True;
			AppendTo[ replBy, {usedDef[[1]], DeleteDuplicates[ condDefs]}],
			newG = g;
			AppendTo[ replBy, {{}, {}}]
		];
		usedForms = {{g}};
		genForms = {{newG}};
		eh = Switch[ n,
				CGB2expandDef,
				{EqualDef$TM, IffDef$TM},
				_,
				{Equal$TM, Iff$TM, EqualDef$TM, IffDef$TM}
			];
		Do[
			(* We do not want to rewrite
				- formulae into the boolean truth constant True (or a formula whose truth value is True),
				- formulae into formulae that are already in the KB, and
				- subexpressions of equalities and equivalences
					TODO:
					This might be problematic if not both elementary- and ground substitutions are activated;
					Good opportunity for prover options
			*)
            {new, usedDef, condDefs} = replaceFirstPossible[ formula@k[[j]], filterRules[ rules, key@k[[j]]], gRules, AcceptTrue -> False, Except -> Join[ k, newK], ExceptHeads -> eh];
            If[ usedDef =!= {},
        		newForm = makeAssumptionFML[ formula -> new];
        		newK = addNewKB[ newK, newForm, Equality -> Strong, Mode -> Append];
                AppendTo[ usedForms, {k[[j]]}];
                AppendTo[ genForms, {newForm}];
				AppendTo[ replBy, {usedDef[[1]], DeleteDuplicates[ condDefs]}];
                rewritten = True,
                (* else: no rewrite in this formula *)
                Block[ {$autoGenerateRules = False}, newK = addNewKB[ newK, k[[j]]]] (* rewrite rules from this formula are already there *)
            ],
            {j, Length[ k]}
        ];
        If[ rewritten,
        	makeANDNODE[ makePRFINFO[ name -> n, used -> usedForms, generated -> genForms, "usedForms" -> replBy], 
				toBeProved[ goal -> newG, kb -> newK, rest]],
			$Failed
        ]
	]


(* ::Section:: *)
(* Rule composition *)

propositionalLogicRules = {"Propositional Logic Rules",
	{goalInKB, True, True, 1, "term"},
	{falseInKB, True, True, 1, "term"},
	{trueGoal, True, True, 1, "term"},
	{contradictionKB, True, True, 1, "term"},
	{CGB2notGoal, False, True, 30},
	{CGB2andGoal, True, True, 6},
	{CGB2andKB, True, False, 4},
	{CGB2orGoal, True, True, 4},
	{CGB2orGoalInteractive, False, True, 4},
	{CGB2orKB, True, True, 19},
	{CGB2implGoalDirect, True, True, 4},
	{equivGoal, True, True, 10},
	{CGB2equivKB, False, True, 15}
	};
	
miscRules = {"Miscellaneous Rules",
	{CGB2exchangeGoalKBInteractive, False, True, 95, "interactive"},
	{CGB2expandNegationGoal, False, True, 29},
	{CGB2expandNegationKB, True, True, 25}
	};
	
integerRules = {"Integer Rules",
	{CGB2integerIntervalKB, True, True, 10},
	{CGB2integerIntervalGoal, True, True, 5}
	};
	
tupleRules = {"Tuple Rules",
	{CGB2subscriptTupleOf, True, True, 65},
	{CGB2elementTupleOf, True, True, 30},
	{CGB2lengthTupleOf, True, True, 30},
	{CGB2nonemptyTupleOf, True, True, 20},
	{CGB2isTupleTupleOf, True, True, 20},
	{CGB2appendElement, True, True, 70},
	{CGB2appendTupleOf, True, True, 69}
	};
	
additionRules = {"Addition Rules",
	{CGB2cancellationGoal, True, True, 5},
	{CGB2cancellationKB, True, True, 5}
	};
	
orderingRules = {"Ordering Rules",
	{CGB2contradictionOrdering, True, True, 2},
	{CGB2notLessEqual, True, True, 4},
	{CGB2reflexiveClosure, False, True, 65},
	{CGB2transitivityKB, True, True, 5},
	{CGB2transitivityGoal, True, True, 65},
	{CGB2transitivityGoalAddition, True, True, 95},
	{CGB2transitivityGoalAdditionLeft, False, True, 95}
	};
	
minmaxRules = {"min/max Rules",
	{CGB2ArgMin, True, True, 19},
	{CGB2ArgMax, True, True, 19},
	{CGB2tupleAdditionGoal, True, True, 4},
	{CGB2minGoalLeft, True, True, 10},
	{CGB2maxGoalRight, True, True, 10},
	{CGB2minKBRight, True, True, 11},
	{CGB2maxKBLeft, True, True, 11}
	};
	
quantifierRules = {"Quantifier Rules",
	{CGB2forallGoal, True, True, 10},
	{CGB2existsKB, True, True, 11},
	{CGB2existsGoal, True, True, 90},
	{CGB2forallKBInteractive, True, True, 81, "interactive"},
	{CGB2existsGoalInteractive, True, True, 80, "interactive"},
	{CGB2existsTupleOfGoal, True, True, 60},
	{CGB2forallTupleOfKB, True, True, 60},
	{CGB2existsAppendGoal, True, True, 60},
	{CGB2forallAppendKB, True, True, 60}
	};
	
implicationRules = {"Implication Rules",
	{CGB2modusPonensTollens, True, True, 25},
	{CGB2selectImplicationInteractive, True, True, 85, "interactive"}
	};
	
rewriteRules = {"Rules for Rewriting",
	{CGB2generateRewriteRules, True, False, 2},
	{CGB2goalRewriting, True, True, 15},
	{CGB2elementarySubstitution, True, True, 7},
	{CGB2groundSubstitution, True, True, 6},
	{CGB2expandDef, True, True, 91}
	};

registerRuleSet[ "Complexity of bivariate Buchberger's algorithm", complexityGBBivariate, {
	propositionalLogicRules,
	miscRules,
	integerRules,
	tupleRules,
	additionRules,
	orderingRules,
	minmaxRules,
	quantifierRules,
	implicationRules,
	rewriteRules
	}]

End[]

EndPackage[]